# -*- coding: utf-8 -*-
"""This module contains the class to compute lensing properties of any elliptical
profile using Shajib (2019)'s Gauss decomposition."""

__author__ = "ajshajib"

import numpy as np
import abc
from scipy.special import comb

from lenstronomy.LensModel.Profiles.gaussian_ellipse_kappa import GaussianEllipseKappa
from lenstronomy.LensModel.Profiles.sersic_utils import SersicUtil
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()

_SQRT_2PI = np.sqrt(2 * np.pi)


@export
class GaussianEllipseKappaSet(LensProfileBase):
    """This class computes the lensing properties of a set of concentric elliptical
    Gaussian convergences."""

    param_names = ["amp", "sigma", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "sigma": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "sigma": 100,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self, use_scipy_wofz=True, min_ellipticity=1e-5):
        """

        :param use_scipy_wofz: To initiate ``class GaussianEllipseKappa``. If ``True``, Gaussian lensing will use ``scipy.special.wofz`` function. Set ``False`` for lower precision, but faster speed.
        :type use_scipy_wofz: ``bool``
        :param min_ellipticity: To be passed to ``class GaussianEllipseKappa``. Minimum ellipticity for Gaussian elliptical lensing calculation. For lower ellipticity than min_ellipticity the equations for the spherical case will be used.
        :type min_ellipticity: ``float``
        """
        self.gaussian_ellipse_kappa = GaussianEllipseKappa(
            use_scipy_wofz=use_scipy_wofz, min_ellipticity=min_ellipticity
        )
        super(GaussianEllipseKappaSet, self).__init__()

    def function(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """Compute the potential function for a set of concentric elliptical Gaussian
        convergence profiles.

        :param x: x coordinate
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param amp: Amplitude of Gaussian, convention: :math:`A/(2 \\pi\\sigma^2) \\exp(-(x^2+y^2/q^2)/2\\sigma^2)`
        :type amp: ``numpy.array`` with ``dtype=float``
        :param sigma: Standard deviation of Gaussian
        :type sigma: ``numpy.array`` with ``dtype=float``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordianate of centroid
        :type center_y: ``float``
        :return: Potential for elliptical Gaussian convergence
        :rtype: ``float``, or ``numpy.array`` with ``shape = x.shape``
        """
        function = np.zeros_like(x, dtype=float)

        for i in range(len(amp)):
            function += self.gaussian_ellipse_kappa.function(
                x, y, amp[i], sigma[i], e1, e2, center_x, center_y
            )
        return function

    def derivatives(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """Compute the derivatives of function angles :math:`\\partial f/\\partial x`,
        :math:`\\partial f/\\partial y` at :math:`x,\\ y` for a set of concentric
        elliptic Gaussian convergence profiles.

        :param x: x coordinate
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param amp: Amplitude of Gaussian, convention: :math:`A/(2 \\pi\\sigma^2) \\exp(-(x^2+y^2/q^2)/2\\sigma^2)`
        :type amp: ``numpy.array`` with ``dtype=float``
        :param sigma: Standard deviation of Gaussian
        :type sigma: ``numpy.array`` with ``dtype=float``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordianate of centroid
        :type center_y: ``float``
        :return: Deflection angle :math:`\\partial f/\\partial x`, :math:`\\partial f/\\partial y` for elliptical Gaussian convergence
        :rtype: tuple ``(float, float)`` or ``(numpy.array, numpy.array)`` with each ``numpy`` array's shape equal to ``x.shape``
        """
        f_x = np.zeros_like(x, dtype=float)
        f_y = np.zeros_like(x, dtype=float)

        for i in range(len(amp)):
            f_x_i, f_y_i = self.gaussian_ellipse_kappa.derivatives(
                x,
                y,
                amp=amp[i],
                sigma=sigma[i],
                e1=e1,
                e2=e2,
                center_x=center_x,
                center_y=center_y,
            )
            f_x += f_x_i
            f_y += f_y_i

        return f_x, f_y

    def hessian(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """Compute Hessian matrix of function :math:`\\partial^2f/\\partial x^2`,
        :math:`\\partial^2 f/\\partial y^2`, :math:`\\partial^2 f/\\partial x\\partial
        y` for a set of concentric elliptic Gaussian convergence profiles.

        :param x: x coordinate
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param amp: Amplitude of Gaussian, convention: :math:`A/(2 \\pi\\sigma^2) \\exp(-(x^2+y^2/q^2)/2\\sigma^2)`
        :type amp: ``numpy.array`` with ``dtype=float``
        :param sigma: Standard deviation of Gaussian
        :type sigma: ``numpy.array`` with ``dtype=float``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordianate of centroid
        :type center_y: ``float``
        :return: Hessian :math:`\\partial^2f/\\partial x^2`, :math:`\\partial^2/\\partial x\\partial y`,
         :math:`\\partial^2/\\partial y\\partial x`, :math:`\\partial^2 f/\\partial y^2` for elliptical Gaussian convergence.
        :rtype: tuple ``(float, float, float)`` , or ``(numpy.array, numpy.array, numpy.array)``
         with each ``numpy`` array's shape equal to ``x.shape``
        """
        f_xx = np.zeros_like(x, dtype=float)
        f_yy = np.zeros_like(x, dtype=float)
        f_xy = np.zeros_like(x, dtype=float)

        for i in range(len(amp)):
            f_xx_i, f_xy_i, _, f_yy_i = self.gaussian_ellipse_kappa.hessian(
                x,
                y,
                amp=amp[i],
                sigma=sigma[i],
                e1=e1,
                e2=e2,
                center_x=center_x,
                center_y=center_y,
            )
            f_xx += f_xx_i
            f_yy += f_yy_i
            f_xy += f_xy_i

        return f_xx, f_xy, f_xy, f_yy

    def density_2d(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """Compute the density of a set of concentric elliptical Gaussian convergence
        profiles :math:`\\sum A/(2\\pi \\sigma^2) \\exp(-( x^2+y^2/q^2)/2\\sigma^2)`.

        :param x: x coordinate
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param amp: Amplitude of Gaussian, convention: :math:`A/(2 \\pi\\sigma^2) \\exp(-(x^2+y^2/q^2)/2\\sigma^2)`
        :type amp: ``numpy.array`` with ``dtype=float``
        :param sigma: Standard deviation of Gaussian
        :type sigma: ``numpy.array`` with ``dtype=float``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordianate of centroid
        :type center_y: ``float``
        :return: Density :math:`\\kappa` for elliptical Gaussian convergence
        :rtype: ``float``, or ``numpy.array`` with shape equal to ``x.shape``
        """
        density_2d = np.zeros_like(x, dtype=float)

        for i in range(len(amp)):
            density_2d += self.gaussian_ellipse_kappa.density_2d(
                x,
                y,
                amp=amp[i],
                sigma=sigma[i],
                e1=e1,
                e2=e2,
                center_x=center_x,
                center_y=center_y,
            )

        return density_2d


@export
class GaussDecompositionAbstract(metaclass=abc.ABCMeta):
    """This abstract class sets up a template for computing lensing properties of an
    elliptical convergence through Shajib (2019)'s Gauss decomposition."""

    def __init__(
        self,
        n_sigma=15,
        sigma_start_mult=0.02,
        sigma_end_mult=15.0,
        precision=10,
        use_scipy_wofz=True,
        min_ellipticity=1e-5,
    ):
        """Set up settings for the Gaussian decomposition. For more details about the
        decomposition parameters, see Shajib (2019).

        :param n_sigma: Number of Gaussian components
        :type n_sigma: ``int``
        :param sigma_start_mult: Lower range of logarithmically spaced sigmas
        :type sigma_start_mult: ``float``
        :param sigma_end_mult: Upper range of logarithmically spaced sigmas
        :type sigma_end_mult: ``float``
        :param precision: Numerical precision of Gaussian decomposition
        :type precision: ``int``
        :param use_scipy_wofz: To be passed to ``class GaussianEllipseKappa``. If ``True``, Gaussian lensing will use ``scipy.special.wofz`` function. Set ``False`` for lower precision, but faster speed.
        :type use_scipy_wofz: ``bool``
        :param min_ellipticity: To be passed to ``class GaussianEllipseKappa``. Minimum ellipticity for Gaussian elliptical lensing calculation. For lower ellipticity than min_ellipticity the equations for the spherical case will be used.
        :type min_ellipticity: ``float``
        """
        self.gaussian_set = GaussianEllipseKappaSet(
            use_scipy_wofz=use_scipy_wofz, min_ellipticity=min_ellipticity
        )

        self.n_sigma = n_sigma
        self.sigma_start_mult = sigma_start_mult
        self.sigma_end_mult = sigma_end_mult
        self.precision = precision

        p = self.precision
        # nodes and weights based on Fourier-Euler method
        # for details Abate & Whitt (2006)
        kes = np.arange(2 * p + 1)
        self.betas = np.sqrt(2 * p * np.log(10) / 3.0 + 2.0 * 1j * np.pi * kes)
        epsilons = np.zeros(2 * p + 1)

        epsilons[0] = 0.5
        epsilons[1 : p + 1] = 1.0
        epsilons[-1] = 1 / 2.0**p

        for k in range(1, p):
            epsilons[2 * p - k] = epsilons[2 * p - k + 1] + 1 / 2.0**p * comb(p, k)

        self.etas = (-1.0) ** kes * epsilons * 10.0 ** (p / 3.0) * 2.0 * _SQRT_2PI

    def gauss_decompose(self, **kwargs):
        r"""Compute the amplitudes and sigmas of Gaussian components using the integral
        transform with Gaussian kernel from Shajib (2019). The returned values are in
        the convention of eq. (2.13).

        :param kwargs: Keyword arguments to send to ``func``
        :return: Amplitudes and standard deviations of the Gaussian components
        :rtype: tuple ``(numpy.array, numpy.array)``
        """
        sigma_start = self.sigma_start_mult * self.get_scale(**kwargs)
        sigma_end = self.sigma_end_mult * self.get_scale(**kwargs)

        sigmas = np.logspace(np.log10(sigma_start), np.log10(sigma_end), self.n_sigma)

        f_sigmas = np.sum(
            self.etas
            * self.get_kappa_1d(
                sigmas[:, np.newaxis] * self.betas[np.newaxis, :], **kwargs
            ).real,
            axis=1,
        )

        del_log_sigma = np.abs(np.diff(np.log(sigmas)).mean())

        amps = f_sigmas * del_log_sigma / _SQRT_2PI

        # weighting for trapezoid method integral
        amps[0] *= 0.5
        amps[-1] *= 0.5

        return amps, sigmas

    @abc.abstractmethod
    def get_scale(self, **kwargs):
        """Abstract method to identify the keyword argument for the scale size among the
        profile parameters of the child class' convergence profile.

        :param kwargs: Keyword arguments
        :return: Scale size
        :rtype: ``float``
        """

    @abc.abstractmethod
    def get_kappa_1d(self, y, **kwargs):
        r"""Abstract method to compute the spherical Sersic profile at y. The concrete
        method has to defined by the child class.

        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param kwargs: Keyword arguments that are defined by the child class that are particular for the convergence profile
        """

    def function(self, x, y, e1=0.0, e2=0.0, center_x=0.0, center_y=0.0, **kwargs):
        r"""Compute the deflection potential of a Gauss-decomposed elliptic convergence.

        :param x: x coordinate
        :type x: ``float``
        :param y: y coordinate
        :type y: ``float``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordinate of centroid
        :type center_y: ``float``
        :param kwargs: Keyword arguments that are defined by the child class that are particular for the convergence profile
        :return: Deflection potential
        :rtype: ``float``
        """
        amps, sigmas = self.gauss_decompose(**kwargs)

        # converting the amplitude convention A -> A/(2*pi*sigma^2)
        amps *= 2.0 * np.pi * sigmas * sigmas

        return self.gaussian_set.function(
            x, y, amps, sigmas, e1, e2, center_x, center_y
        )

    def derivatives(self, x, y, e1=0.0, e2=0.0, center_x=0.0, center_y=0.0, **kwargs):
        r"""Compute the derivatives of the deflection potential :math:`\partial
        f/\partial x`, :math:`\partial f/\partial y` for a Gauss-decomposed elliptic
        convergence.

        :param x: x coordinate
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordinate of centroid
        :type center_y: ``float``
        :param kwargs: Keyword arguments that are defined by the child class that are particular for the convergence profile
        :return: Derivatives of deflection potential
        :rtype: tuple ``(type(x), type(x))``
        """
        amps, sigmas = self.gauss_decompose(**kwargs)

        # converting the amplitude convention A -> A/(2*pi*sigma^2)
        amps *= 2.0 * np.pi * sigmas * sigmas

        return self.gaussian_set.derivatives(
            x, y, amps, sigmas, e1, e2, center_x, center_y
        )

    def hessian(self, x, y, e1=0.0, e2=0.0, center_x=0.0, center_y=0.0, **kwargs):
        r"""Compute the Hessian of the deflection potential :math:`\partial^2f/\partial
        x^2`, :math:`\partial^2 f/ \partial y^2`, :math:`\partial^2 f/\partial x\partial
        y` of a Gauss-decomposed elliptic Sersic convergence.

        :param x: x coordinate
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordinate of centroid
        :type center_y: ``float``
        :param kwargs: Keyword arguments that are defined by the child class that are particular for the convergence profile
        :return: Hessian of deflection potential
        :rtype: tuple ``(type(x), type(x), type(x))``
        """
        amps, sigmas = self.gauss_decompose(**kwargs)

        # converting the amplitude convention A -> A/(2*pi*sigma^2)
        amps *= 2.0 * np.pi * sigmas * sigmas

        return self.gaussian_set.hessian(x, y, amps, sigmas, e1, e2, center_x, center_y)

    def density_2d(self, x, y, e1=0.0, e2=0.0, center_x=0.0, center_y=0.0, **kwargs):
        r"""Compute the convergence profile for Gauss-decomposed elliptic Sersic profile.

        :param x: x coordinate
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordinate of centroid
        :type center_y: ``float``
        :param kwargs: Keyword arguments that are defined by the child class that are particular for the convergence profile in the child class.
        :return: Convergence profile
        :rtype: ``type(x)``
        """
        amps, sigmas = self.gauss_decompose(**kwargs)

        # converting the amplitude convention A -> A/(2*pi*sigma^2)
        amps *= 2.0 * np.pi * sigmas * sigmas

        return self.gaussian_set.density_2d(
            x, y, amps, sigmas, e1, e2, center_x, center_y
        )


@export
class SersicEllipseGaussDec(GaussDecompositionAbstract):
    """This class computes the lensing properties of an elliptical Sersic profile using
    the Shajib (2019)'s Gauss decomposition method."""

    param_names = ["k_eff", "R_sersic", "n_sersic", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "k_eff": 0.0,
        "R_sersic": 0.0,
        "n_sersic": 0.5,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100.0,
        "center_y": -100.0,
    }
    upper_limit_default = {
        "k_eff": 100.0,
        "R_sersic": 100.0,
        "n_sersic": 8.0,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100.0,
        "center_y": 100.0,
    }

    def get_kappa_1d(self, y, **kwargs):
        r"""Compute the spherical Sersic profile at y.

        :param y: y coordinate
        :type y: ``float``
        :param kwargs: Keyword arguments

        :Keyword Arguments:
            * **n_sersic** (``float``) --
              Sersic index
            * **R_sersic** (``float``) --
              Sersic scale radius
            * **k_eff** (``float``) --
              Sersic convergence at R_sersic

        :return: Sersic function at y
        :rtype: ``type(y)``
        """
        n_sersic = kwargs["n_sersic"]
        R_sersic = kwargs["R_sersic"]
        k_eff = kwargs["k_eff"]

        bn = SersicUtil.b_n(n_sersic)

        return k_eff * np.exp(-bn * (y / R_sersic) ** (1.0 / n_sersic) + bn)

    def get_scale(self, **kwargs):
        """Identify the scale size from the keyword arguments.

        :param kwargs: Keyword arguments

        :Keyword Arguments:
            * **n_sersic** (``float``) --
              Sersic index
            * **R_sersic** (``float``) --
              Sersic scale radius
            * **k_eff** (``float``) --
              Sersic convergence at R_sersic

        :return: Sersic radius
        :rtype: ``float``
        """
        return kwargs["R_sersic"]


@export
class NFWEllipseGaussDec(GaussDecompositionAbstract):
    """This class computes the lensing properties of an elliptical, projected NFW
    profile using Shajib (2019)'s Gauss decomposition method."""

    param_names = ["Rs", "alpha_Rs", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "Rs": 0,
        "alpha_Rs": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "Rs": 100,
        "alpha_Rs": 10,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(
        self,
        n_sigma=15,
        sigma_start_mult=0.005,
        sigma_end_mult=50.0,
        precision=10,
        use_scipy_wofz=True,
        min_ellipticity=1e-5,
    ):
        """Set up settings for the Gaussian decomposition. For more details about the
        decomposition parameters, see Shajib (2019).

        :param n_sigma: Number of Gaussian components
        :type n_sigma: ``int``
        :param sigma_start_mult: Lower range of logarithmically spaced sigmas
        :type sigma_start_mult: ``float``
        :param sigma_end_mult: Upper range of logarithmically spaced sigmas
        :type sigma_end_mult: ``float``
        :param precision: Numerical precision of Gaussian decomposition
        :type precision: ``int``
        :param use_scipy_wofz: To be passed to ``class GaussianEllipseKappa``. If ``True``, Gaussian lensing will use ``scipy.special.wofz`` function. Set ``False`` for lower precision, but faster speed.
        :type use_scipy_wofz: ``bool``
        :param min_ellipticity: To be passed to ``class GaussianEllipseKappa``. Minimum ellipticity for Gaussian elliptical lensing calculation. For lower ellipticity than min_ellipticity the equations for the spherical case will be used.
        :type min_ellipticity: ``float``
        """
        super(NFWEllipseGaussDec, self).__init__(
            n_sigma=n_sigma,
            sigma_start_mult=sigma_start_mult,
            sigma_end_mult=sigma_end_mult,
            precision=precision,
            use_scipy_wofz=use_scipy_wofz,
            min_ellipticity=min_ellipticity,
        )

    def get_kappa_1d(self, y, **kwargs):
        r"""Compute the spherical projected NFW profile at y.

        :param y: y coordinate
        :type y: ``float``
        :param kwargs: Keyword arguments

        :Keyword Arguments:
            * **alpha_Rs** (``float``) --
              Deflection angle at ``Rs``
            * **R_s** (``float``) --
              NFW scale radius

        :return: projected NFW profile at y
        :rtype: ``type(y)``
        """
        R_s = kwargs["Rs"]
        alpha_Rs = kwargs["alpha_Rs"]

        kappa_s = alpha_Rs / (4 * R_s * (1 - 0.30102999566))
        # log2 = 0.30102999566

        x = y / R_s

        f = np.empty(shape=x.shape, dtype=x.dtype)

        range1 = x > 1.0
        if np.any(range1):
            s = x[range1]
            f[range1] = (1 - np.arccos(1 / s) / np.sqrt(s * s - 1)) / (s * s - 1)

        range2 = x < 1.0
        if np.any(range2):
            s = x[range2]
            f[range2] = (1 - np.arccosh(1 / s) / np.sqrt(1 - s * s)) / (s * s - 1)

        range3 = np.logical_and(np.logical_not(range1), np.logical_not(range2))
        if np.any(range3):
            f[range3] = 1.0 / 3.0

        return 2 * kappa_s * f

    def get_scale(self, **kwargs):
        """Identify the scale size from the keyword arguments.

        :param kwargs: Keyword arguments

        :Keyword Arguments:
            * **alpha_Rs** (``float``) --
              Deflection angle at ``Rs``
            * **R_s** (``float``) --
              NFW scale radius

        :return: NFW scale radius
        :rtype: ``float``
        """
        return kwargs["Rs"]


@export
class GaussDecompositionAbstract3D(GaussDecompositionAbstract):
    """This abstract class sets up a template for computing lensing properties of a
    convergence from 3D spherical profile through Shajib (2019)'s Gauss
    decomposition."""

    def gauss_decompose(self, **kwargs):
        r"""Compute the amplitudes and sigmas of Gaussian components using the integral
        transform with Gaussian kernel from Shajib (2019). The returned values are in
        the convention of eq. (2.13).

        :param kwargs: Keyword arguments to send to ``func``
        :return: Amplitudes and standard deviations of the Gaussian components
        :rtype: tuple ``(numpy.array, numpy.array)``
        """

        f_sigmas, sigmas = super(GaussDecompositionAbstract3D, self).gauss_decompose(
            **kwargs
        )

        return f_sigmas * sigmas * _SQRT_2PI, sigmas


@export
class CTNFWGaussDec(GaussDecompositionAbstract3D):
    """This class computes the lensing properties of an projection from a spherical
    cored-truncated NFW profile using Shajib (2019)'s Gauss decomposition method."""

    param_names = ["r_s", "r_core", "r_trunc", "a", "rho_s", "center_x" "center_y"]
    lower_limit_default = {
        "r_s": 0,
        "r_core": 0,
        "r_trunc": 0,
        "a": 0.0,
        "rho_s": 0,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "r_s": 100,
        "r_core": 100,
        "r_trunc": 100,
        "a": 10.0,
        "rho_s": 1000,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(
        self,
        n_sigma=15,
        sigma_start_mult=0.01,
        sigma_end_mult=20.0,
        precision=10,
        use_scipy_wofz=True,
    ):
        """Set up settings for the Gaussian decomposition. For more details about the
        decomposition parameters, see Shajib (2019).

        :param n_sigma: Number of Gaussian components
        :type n_sigma: ``int``
        :param sigma_start_mult: Lower range of logarithmically spaced sigmas
        :type sigma_start_mult: ``float``
        :param sigma_end_mult: Upper range of logarithmically spaced sigmas
        :type sigma_end_mult: ``float``
        :param precision: Numerical precision of Gaussian decomposition
        :type precision: ``int``
        :param use_scipy_wofz: To be passed to ``class GaussianEllipseKappa``. If ``True``, Gaussian lensing will use ``scipy.special.wofz`` function. Set ``False`` for lower precision, but faster speed.
        :type use_scipy_wofz: ``bool``
        """
        super(CTNFWGaussDec, self).__init__(
            n_sigma=n_sigma,
            sigma_start_mult=sigma_start_mult,
            sigma_end_mult=sigma_end_mult,
            precision=precision,
            use_scipy_wofz=use_scipy_wofz,
        )

    def get_kappa_1d(self, y, **kwargs):
        r"""Compute the spherical cored-truncated NFW profile at y.

        :param y: y coordinate
        :type y: ``float``
        :param kwargs: Keyword arguments

        :Keyword Arguments:
            * **r_s** (``float``) --
              Scale radius
            * **r_trunc** (``float``) --
              Truncation radius
            * **r_core** (``float``) --
              Core radius
            * **rho_s** (``float``) --
              Density normalization
            * **a** (``float``) --
              Core regularization parameter

        :return: projected NFW profile at y
        :rtype: ``type(y)``
        """
        r_s = kwargs["r_s"]
        r_trunc = kwargs["r_trunc"]
        r_core = kwargs["r_core"]
        rho_s = kwargs["rho_s"]
        a = kwargs["a"]

        beta = r_core / r_s
        tau = r_trunc / r_s

        x = y / r_s

        return (
            rho_s
            * (tau * tau / (tau * tau + x * x))
            / (x**a + beta**a) ** (1.0 / a)
            / (1.0 + x) ** 2
        )

    def get_scale(self, **kwargs):
        """Identify the scale size from the keyword arguments.

        :param kwargs: Keyword arguments

        :Keyword Arguments:
            * **r_s** (``float``) --
              Scale radius
            * **r_trunc** (``float``) --
              Truncation radius
            * **r_core** (``float``) --
              Core radius
            * **rho_s** (``float``) --
              Density normalization
            * **a** (``float``) --
              Core regularization parameter

        :return: NFW scale radius
        :rtype: ``float``
        """
        return kwargs["r_s"]



import numpy as np
import numpy.testing as npt
import pytest
import matplotlib.pyplot as plt
from lenstronomy.Data.coord_transforms import Coordinates

import lenstronomy.Plots.plot_util as plot_util


class TestPlotUtil(object):
    def setup_method(self):
        pass

    def test_sqrt(self):
        image = np.random.randn(10, 10)
        image_rescaled = plot_util.sqrt(image)
        npt.assert_almost_equal(np.min(image_rescaled), 0)

    def test_scale_bar(self):
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        plot_util.scale_bar(ax, 3, dist=1, text='1"', flipped=True)
        plt.close()
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        plot_util.text_description(
            ax, d=3, text="test", color="w", backgroundcolor="k", flipped=True
        )
        plt.close()

    def test_source_position_plot(self):
        from lenstronomy.PointSource.point_source import PointSource
        from lenstronomy.LensModel.lens_model import LensModel

        lensModel = LensModel(lens_model_list=["SIS"])
        ps = PointSource(
            point_source_type_list=["UNLENSED", "LENSED_POSITION", "SOURCE_POSITION"],
            lensModel=lensModel,
        )
        kwargs_lens = [{"theta_E": 1.0, "center_x": 0, "center_y": 0}]
        kwargs_ps = [
            {"ra_image": [1.0, 1.0], "dec_image": [0, 1], "point_amp": [1, 1]},
            {"ra_image": [1.0], "dec_image": [1.0], "point_amp": [10]},
            {"ra_source": 0.1, "dec_source": 0, "point_amp": 1.0},
        ]
        ra_source, dec_source = ps.source_position(kwargs_ps, kwargs_lens)
        from lenstronomy.Data.coord_transforms import Coordinates

        coords_source = Coordinates(
            transform_pix2angle=np.array([[1, 0], [0, 1]]) * 0.1,
            ra_at_xy_0=-2,
            dec_at_xy_0=-2,
        )

        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        plot_util.source_position_plot(ax, coords_source, ra_source, dec_source)
        plt.close()

    def test_result_string(self):
        x = np.random.normal(loc=1, scale=0.1, size=10000)
        string = plot_util.result_string(x, weights=None, title_fmt=".2f", label="test")
        print(string)
        assert string == str("test = ${1.00}_{-0.10}^{+0.10}$")

    def test_cmap_conf(self):
        cmap = plot_util.cmap_conf(cmap_string="gist_heat")
        cmap_update = plot_util.cmap_conf(cmap_string=cmap)
        assert cmap.name == cmap_update.name

    def test_plot_line_set(self):
        coords = Coordinates(
            transform_pix2angle=[[1, 0], [0, 1]], ra_at_xy_0=0, dec_at_xy_0=0
        )
        line_set_x = np.linspace(start=0, stop=1, num=10)
        line_set_y = np.linspace(start=0, stop=1, num=10)
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        ax = plot_util.plot_line_set(
            ax,
            coords,
            line_set_x,
            line_set_y,
            origin=None,
            color="g",
            flipped_x=True,
            pixel_offset=False,
        )
        plt.close()

        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        ax = plot_util.plot_line_set(
            ax,
            coords,
            line_set_x,
            line_set_y,
            origin=[1, 1],
            color="g",
            flipped_x=False,
            pixel_offset=True,
        )
        plt.close()

        # and here we input a list of arrays

        line_set_list_x = [
            np.linspace(start=0, stop=1, num=10),
            np.linspace(start=0, stop=1, num=10),
        ]
        line_set_list_y = [
            np.linspace(start=0, stop=1, num=10),
            np.linspace(start=0, stop=1, num=10),
        ]
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        ax = plot_util.plot_line_set(
            ax,
            coords,
            line_set_list_x,
            line_set_list_y,
            origin=None,
            color="g",
            flipped_x=True,
        )
        plt.close()

        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        ax = plot_util.plot_line_set(
            ax,
            coords,
            line_set_list_x,
            line_set_list_y,
            origin=[1, 1],
            color="g",
            flipped_x=False,
        )
        plt.close()

    def test_image_position_plot(self):
        coords = Coordinates(
            transform_pix2angle=[[1, 0], [0, 1]], ra_at_xy_0=0, dec_at_xy_0=0
        )
        f, ax = plt.subplots(1, 1, figsize=(4, 4))

        ra_image, dec_image = np.array([1, 2]), np.array([1, 2])
        ax = plot_util.image_position_plot(
            ax,
            coords,
            ra_image,
            dec_image,
            color="w",
            image_name_list=None,
            origin=None,
            flipped_x=False,
            pixel_offset=False,
        )
        plt.close()
        ax = plot_util.image_position_plot(
            ax,
            coords,
            ra_image,
            dec_image,
            color="w",
            image_name_list=["A", "B"],
            origin=[1, 1],
            flipped_x=True,
            pixel_offset=True,
        )
        plt.close()

    def test_cmap_copy(self):
        from lenstronomy.Plots.plot_util import cmap_conf

        cmap_new = cmap_conf("gist_heat")


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

# this file contains a class to make an Ellipsoid
import numpy as np
from lenstronomy.Util import param_util

__all__ = ["Ellipsoid"]


class Ellipsoid(object):
    """Class for an universal surface brightness within an ellipsoid."""

    def __init__(self):
        self.param_names = ["amp", "radius", "e1", "e2", "center_x", "center_y"]
        self.lower_limit_default = {
            "amp": 0,
            "radius": 0,
            "e1": -0.5,
            "e2": -0.5,
            "center_x": -100,
            "center_y": -100,
        }
        self.upper_limit_default = {
            "amp": 1000,
            "radius": 100,
            "e1": 0.5,
            "e2": 0.5,
            "center_x": 100,
            "center_y": 100,
        }

    def function(self, x, y, amp, radius, e1, e2, center_x, center_y):
        """

        :param x:
        :param y:
        :param amp: surface brightness within the ellipsoid
        :param radius: radius (product average of semi-major and semi-minor axis) of the ellipsoid
        :param e1: eccentricity
        :param e2: eccentricity
        :param center_x: center
        :param center_y: center
        :return: surface brightness
        """
        x_, y_ = param_util.transform_e1e2_product_average(
            x, y, e1, e2, center_x, center_y
        )
        r2 = x_**2 + y_**2
        flux = np.zeros_like(x)
        flux[r2 <= radius**2] = 1
        area = np.pi * radius**2
        return amp / area * flux


def function(x, y, amp, sigma, center_x, center_y):
    """Returns torus (ellipse with constant surface brightness) profile."""
    x_shift = x - center_x
    y_shift = y - center_y
    area = np.pi * sigma**2
    dist = (x_shift / sigma) ** 2 + (y_shift / sigma) ** 2
    torus = np.zeros_like(x)
    torus[dist <= 1] = 1
    return amp / area * torus



__author__ = "sibirrer"

from lenstronomy.LensModel.Profiles.nfw import NFW
from lenstronomy.LensModel.Profiles.nfw_ellipse_cse import NFW_ELLIPSE_CSE
from lenstronomy.Analysis.lens_profile import LensProfileAnalysis
from lenstronomy.LensModel.lens_model import LensModel

import numpy as np
import numpy.testing as npt
import pytest


class TestNFWELLIPSE(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.nfw = NFW()
        self.nfw_cse = NFW_ELLIPSE_CSE(high_accuracy=True)
        self.nfw_cse_low_accuracy = NFW_ELLIPSE_CSE(high_accuracy=False)

    def test_function(self):
        x = np.linspace(0.01, 2, 10)
        y = np.zeros_like(x)
        kwargs = {"alpha_Rs": 2, "Rs": 2, "center_x": 0, "center_y": 0}

        f_nfw = self.nfw.function(x, y, **kwargs)
        f_cse = self.nfw_cse.function(x, y, e1=0, e2=0, **kwargs)
        npt.assert_almost_equal(f_cse, f_nfw, decimal=5)
        f_cse_low = self.nfw_cse_low_accuracy.function(x, y, e1=0, e2=0, **kwargs)
        npt.assert_almost_equal(f_cse_low / f_nfw, 1, decimal=3)

    def test_derivatives(self):
        x = np.linspace(0.01, 2, 10)
        y = np.zeros_like(x)
        kwargs = {"alpha_Rs": 0.5, "Rs": 2, "center_x": 0, "center_y": 0}

        f_x_nfw, f_y_nfw = self.nfw.derivatives(x, y, **kwargs)
        f_x_cse, f_y_cse = self.nfw_cse.derivatives(x, y, e1=0, e2=0, **kwargs)
        npt.assert_almost_equal(f_x_cse, f_x_nfw, decimal=5)
        npt.assert_almost_equal(f_y_cse, f_y_nfw, decimal=5)
        f_x_cse_low, f_y_cse_low = self.nfw_cse_low_accuracy.derivatives(
            x, y, e1=0, e2=0, **kwargs
        )
        npt.assert_almost_equal(f_x_cse_low / f_x_nfw, 1, decimal=2)
        npt.assert_almost_equal(f_y_cse_low, f_y_nfw, decimal=2)

    def test_hessian(self):
        x = np.linspace(0.01, 2, 10)
        y = np.zeros_like(x)
        kwargs = {"alpha_Rs": 0.5, "Rs": 2, "center_x": 0, "center_y": 0}

        f_xx_nfw, f_xy_nfw, f_yx_nfw, f_yy_nfw = self.nfw.hessian(x, y, **kwargs)
        f_xx_cse, f_xy_cse, f_yx_cse, f_yy_cse = self.nfw_cse.hessian(
            x, y, e1=0, e2=0, **kwargs
        )
        npt.assert_almost_equal(f_xx_cse, f_xx_nfw, decimal=5)
        npt.assert_almost_equal(f_xy_cse, f_xy_nfw, decimal=5)
        npt.assert_almost_equal(f_yx_cse, f_yx_nfw, decimal=5)
        npt.assert_almost_equal(f_yy_cse, f_yy_nfw, decimal=5)

        f_xx_cse, f_xy_cse, f_yx_cse, f_yy_cse = self.nfw_cse_low_accuracy.hessian(
            x, y, e1=0, e2=0, **kwargs
        )
        npt.assert_almost_equal(f_xx_cse / f_xx_nfw, 1, decimal=1)
        npt.assert_almost_equal(f_xy_cse, f_xy_nfw, decimal=5)
        npt.assert_almost_equal(f_yx_cse, f_yx_nfw, decimal=5)
        npt.assert_almost_equal(f_yy_cse / f_yy_nfw, 1, decimal=1)

    def test_mass_3d_lens(self):
        R = 1
        Rs = 3
        alpha_Rs = 1
        m_3d_nfw = self.nfw.mass_3d_lens(R, Rs, alpha_Rs)
        m_3d_cse = self.nfw_cse.mass_3d_lens(R, Rs, alpha_Rs)
        npt.assert_almost_equal(m_3d_nfw, m_3d_cse, decimal=8)

    def test_ellipticity(self):
        """Test the definition of the ellipticity normalization (along major axis or
        product averaged axes)"""
        x, y = np.linspace(start=0.001, stop=10, num=100), np.zeros(100)
        kwargs_round = {
            "alpha_Rs": 0.5,
            "Rs": 2,
            "center_x": 0,
            "center_y": 0,
            "e1": 0,
            "e2": 0,
        }
        kwargs = {
            "alpha_Rs": 0.5,
            "Rs": 2,
            "center_x": 0,
            "center_y": 0,
            "e1": 0.3,
            "e2": 0,
        }

        f_xx, f_xy, f_yx, f_yy = self.nfw_cse.hessian(x, y, **kwargs_round)
        kappa_round = 1.0 / 2 * (f_xx + f_yy)

        f_xx, f_xy, f_yx, f_yy = self.nfw_cse.hessian(x, y, **kwargs)
        kappa_major = 1.0 / 2 * (f_xx + f_yy)

        f_xx, f_xy, f_yx, f_yy = self.nfw_cse.hessian(y, x, **kwargs)
        kappa_minor = 1.0 / 2 * (f_xx + f_yy)

        npt.assert_almost_equal(
            np.sqrt(kappa_minor * kappa_major), kappa_round, decimal=2
        )

        # import matplotlib.pyplot as plt
        # plt.plot(x, kappa_round/kappa_round, ':', label='round', alpha=0.5)
        # plt.plot(x, kappa_major/kappa_round, ',-', label='major', alpha=0.5)
        # plt.plot(x, kappa_minor/kappa_round, '--', label='minor', alpha=0.5)
        # plt.plot(x, np.sqrt(kappa_minor * kappa_major)/kappa_round, '--', label='prod', alpha=0.5)
        # plt.plot(x, np.sqrt(kappa_minor**2 + kappa_major**2) / kappa_round / 2, '--', label='square', alpha=0.5)
        # plt.legend()
        # plt.show()

    def test_einstein_rad(self):
        """Test that the Einstein radius doesn't change significantly with
        ellipticity."""
        kwargs_round = {
            "alpha_Rs": 0.5,
            "Rs": 2,
            "center_x": 0,
            "center_y": 0,
            "e1": 0,
            "e2": 0,
        }
        kwargs = {
            "alpha_Rs": 0.5,
            "Rs": 2,
            "center_x": 0,
            "center_y": 0,
            "e1": 0.3,
            "e2": 0,
        }
        LensMod = LensModel(["NFW_ELLIPSE_CSE"])
        LensAn = LensProfileAnalysis(LensMod)
        r_Ein_round = LensAn.effective_einstein_radius([kwargs_round])
        r_Ein_ell = LensAn.effective_einstein_radius([kwargs])
        npt.assert_almost_equal(r_Ein_round, r_Ein_ell, decimal=1)


if __name__ == "__main__":
    pytest.main()



from lenstronomy.LensModel.Profiles.hernquist_ellipse import Hernquist_Ellipse
import lenstronomy.Util.param_util as param_util
from lenstronomy.Util import util
from lenstronomy.LensModel.Profiles.cored_steep_ellipsoid import CSEMajorAxisSet
import numpy as np

__all__ = ["HernquistEllipseCSE"]


class HernquistEllipseCSE(Hernquist_Ellipse):
    """This class contains functions for the elliptical Hernquist profile.

    Ellipticity is defined in the convergence.
    Approximation with CSE profile introduced by Oguri 2021: https://arxiv.org/pdf/2106.11464.pdf
    """

    param_names = ["sigma0", "Rs", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "sigma0": 0,
        "Rs": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "sigma0": 100,
        "Rs": 100,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.cse_major_axis_set = CSEMajorAxisSet()
        # Table 2 in Oguri 2021
        self._a_list = [
            9.200445e-18,
            2.184724e-16,
            3.548079e-15,
            2.823716e-14,
            1.091876e-13,
            6.998697e-13,
            3.142264e-12,
            1.457280e-11,
            4.472783e-11,
            2.042079e-10,
            8.708137e-10,
            2.423649e-09,
            7.353440e-09,
            5.470738e-08,
            2.445878e-07,
            4.541672e-07,
            3.227611e-06,
            1.110690e-05,
            3.725101e-05,
            1.056271e-04,
            6.531501e-04,
            2.121330e-03,
            8.285518e-03,
            4.084190e-02,
            5.760942e-02,
            1.788945e-01,
            2.092774e-01,
            3.697750e-01,
            3.440555e-01,
            5.792737e-01,
            2.325935e-01,
            5.227961e-01,
            3.079968e-01,
            1.633456e-01,
            7.410900e-02,
            3.123329e-02,
            1.292488e-02,
            2.156527e00,
            1.652553e-02,
            2.314934e-02,
            3.992313e-01,
        ]
        self._s_list = [
            1.199110e-06,
            3.751762e-06,
            9.927207e-06,
            2.206076e-05,
            3.781528e-05,
            6.659808e-05,
            1.154366e-04,
            1.924150e-04,
            3.040440e-04,
            4.683051e-04,
            7.745084e-04,
            1.175953e-03,
            1.675459e-03,
            2.801948e-03,
            9.712807e-03,
            5.469589e-03,
            1.104654e-02,
            1.893893e-02,
            2.792864e-02,
            4.152834e-02,
            6.640398e-02,
            1.107083e-01,
            1.648028e-01,
            2.839601e-01,
            4.129439e-01,
            8.239115e-01,
            6.031726e-01,
            1.145604e00,
            1.401895e00,
            2.512223e00,
            2.038025e00,
            4.644014e00,
            9.301590e00,
            2.039273e01,
            4.896534e01,
            1.252311e02,
            3.576766e02,
            2.579464e04,
            2.944679e04,
            2.834717e03,
            5.931328e04,
        ]
        super(HernquistEllipseCSE, self).__init__()

    def function(self, x, y, sigma0, Rs, e1, e2, center_x=0, center_y=0):
        """Returns double integral of NFW profile."""
        phi_q, q = param_util.ellipticity2phi_q(e1, e2)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_q)

        # potential calculation
        f_ = self.cse_major_axis_set.function(
            x__ / Rs, y__ / Rs, self._a_list, self._s_list, q
        )
        const = self._normalization(sigma0, Rs, q)
        return const * f_

    def derivatives(self, x, y, sigma0, Rs, e1, e2, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function (integral of NFW)"""
        phi_q, q = param_util.ellipticity2phi_q(e1, e2)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_q)
        f__x, f__y = self.cse_major_axis_set.derivatives(
            x__ / Rs, y__ / Rs, self._a_list, self._s_list, q
        )

        # rotate deflections back
        f_x, f_y = util.rotate(f__x, f__y, -phi_q)
        const = self._normalization(sigma0, Rs, q) / Rs
        return const * f_x, const * f_y

    def hessian(self, x, y, sigma0, Rs, e1, e2, center_x=0, center_y=0):
        """Returns Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx,
        d^f/dy^2."""
        phi_q, q = param_util.ellipticity2phi_q(e1, e2)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_q)
        f__xx, f__xy, __, f__yy = self.cse_major_axis_set.hessian(
            x__ / Rs, y__ / Rs, self._a_list, self._s_list, q
        )

        # rotate back
        kappa = 1.0 / 2 * (f__xx + f__yy)
        gamma1__ = 1.0 / 2 * (f__xx - f__yy)
        gamma2__ = f__xy
        gamma1 = np.cos(2 * phi_q) * gamma1__ - np.sin(2 * phi_q) * gamma2__
        gamma2 = +np.sin(2 * phi_q) * gamma1__ + np.cos(2 * phi_q) * gamma2__
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        const = self._normalization(sigma0, Rs, q) / Rs**2

        return const * f_xx, const * f_xy, const * f_xy, const * f_yy

    @staticmethod
    def _normalization(sigma0, Rs, q):
        """Mapping to eqn 10 and 11 in Oguri 2021 from phenomenological definition.

        :param sigma0: sigma0 normalization
        :param Rs: scale radius
        :param q: axis ratio
        :return: normalization (m)
        """
        rs_ = Rs / np.sqrt(q)
        const = sigma0 / 2 * rs_**3
        return const



import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.spp import SPP
import numpy as np
import scipy.special as special

__all__ = ["PowerLaw"]


class PowerLaw(object):
    """Class for power-law elliptical light distribution."""

    param_names = ["amp", "gamma", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "gamma": 1,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "gamma": 3,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.lens = SPP()

    def function(self, x, y, amp, gamma, e1, e2, center_x=0, center_y=0):
        """

        :param x: ra-coordinate
        :param y: dec-coordinate
        :param amp: amplitude of flux
        :param gamma: projected power-law slope
        :param e1: ellipticity
        :param e2: ellipticity
        :param center_x: center
        :param center_y: center
        :return: projected flux
        """
        x_, y_ = param_util.transform_e1e2_product_average(
            x, y, e1, e2, center_x, center_y
        )
        _s = 0.0001
        a = x_**2 + y_**2 + _s**2

        sigma = amp * a ** ((1.0 - gamma) / 2.0)
        return sigma

    def light_3d(self, r, amp, gamma, e1=0, e2=0):
        """

        :param r:
        :param amp:
        :param gamma:
        :param e1:
        :param e2:
        :return:
        """
        rho0 = self._amp2rho(amp, gamma)
        rho = rho0 / r**gamma
        return rho

    @staticmethod
    def _amp2rho(amp, gamma):
        """

        :param amp:
        :param gamma:
        :return:
        """
        factor = (
            np.sqrt(np.pi)
            * special.gamma(1.0 / 2 * (-1 + gamma))
            / special.gamma(gamma / 2.0)
        )
        return amp / factor



__author__ = "sibirrer"

from scipy import stats
import numpy as np

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
class SkewGaussian(object):
    """Class for the Skew Gaussian distribution."""

    def pdf(self, x, e=0.0, w=1.0, a=0.0):
        """
        probability density function
        see: https://en.wikipedia.org/wiki/Skew_normal_distribution

        :param x: input value
        :param e:
        :param w:
        :param a:
        :return:
        """
        t = (x - e) / w
        return 2.0 / w * stats.norm.pdf(t) * stats.norm.cdf(a * t)

    def pdf_skew(self, x, mu, sigma, skw):
        """Function with different parameterisation.

        :param x:
        :param mu: mean
        :param sigma: sigma
        :param skw: skewness
        :return:
        """
        if skw >= 1 or skw <= -1:
            raise ValueError("skewness %s out of range" % skw)
        e, w, a = self.map_mu_sigma_skw(mu, sigma, skw)
        pdf = self.pdf(x, e, w, a)
        return pdf

    def _delta_skw(self, skw):
        """

        :param skw: skewness parameter
        :return: delta
        """
        skw_23 = np.abs(skw) ** (2.0 / 3)
        delta2 = skw_23 * np.pi / 2 / (skw_23 + ((4 - np.pi) / 2) ** (2.0 / 3))
        return np.sqrt(delta2) * skw / np.abs(skw)

    def _alpha_delta(self, delta):
        """

        :param delta: delta parameter
        :return: alpha (a)
        """
        return delta / np.sqrt(1 - delta**2)

    def _w_sigma_delta(self, sigma, delta):
        """Invert variance.

        :param sigma:
        :param delta:
        :return: w parameter
        """
        sigma2 = sigma**2
        w2 = sigma2 / (1 - 2 * delta**2 / np.pi)
        w = np.sqrt(w2)
        return w

    def _e_mu_w_delta(self, mu, w, delta):
        """

        :param mu:
        :param w:
        :param delta:
        :return: epsilon (e)
        """
        e = mu - w * delta * np.sqrt(2 / np.pi)
        return e

    def map_mu_sigma_skw(self, mu, sigma, skw):
        """Map to parameters e, w, a.

        :param mu: mean
        :param sigma: standard deviation
        :param skw: skewness
        :return: e, w, a
        """
        delta = self._delta_skw(skw)
        a = self._alpha_delta(delta)
        w = self._w_sigma_delta(sigma, delta)
        e = self._e_mu_w_delta(mu, w, delta)
        return e, w, a


@export
class KDE1D(object):
    """Class that allows to compute likelihoods based on a 1-d posterior sample."""

    def __init__(self, values):
        """

        :param values: 1d numpy array of points representing a PDF
        """
        self._points = values
        self._kernel = stats.gaussian_kde(values)

    def likelihood(self, x):
        """

        :param x: position where to evaluate the density
        :return: likelihood given the sample distribution
        """

        dens = self._kernel.evaluate(points=x)
        return dens


@export
def compute_lower_upper_errors(sample, num_sigma=1):
    """Computes the upper and lower sigma from the median value. This functions gives
    good error estimates for skewed pdf's.

    :param sample: 1-D sample
    :param num_sigma: integer, number of sigmas to be returned
    :return: median, lower_sigma, upper_sigma
    """
    if num_sigma > 3:
        raise ValueError(
            "Number of sigma-constraints restricted to three. %s not valid" % num_sigma
        )
    num = len(sample)
    num_threshold1 = int(round((num - 1) * 0.841345))
    num_threshold2 = int(round((num - 1) * 0.977249868))
    num_threshold3 = int(round((num - 1) * 0.998650102))

    median = np.median(sample)
    sorted_sample = np.sort(sample)
    if num_sigma > 0:
        upper_sigma1 = sorted_sample[num_threshold1 - 1]
        lower_sigma1 = sorted_sample[num - num_threshold1 - 1]
    else:
        return median, [[]]
    if num_sigma > 1:
        upper_sigma2 = sorted_sample[num_threshold2 - 1]
        lower_sigma2 = sorted_sample[num - num_threshold2 - 1]
    else:
        return median, [[median - lower_sigma1, upper_sigma1 - median]]
    if num_sigma > 2:
        upper_sigma3 = sorted_sample[num_threshold3 - 1]
        lower_sigma3 = sorted_sample[num - num_threshold3 - 1]
        return median, [
            [median - lower_sigma1, upper_sigma1 - median],
            [median - lower_sigma2, upper_sigma2 - median],
            [median - lower_sigma3, upper_sigma3 - median],
        ]
    else:
        return median, [
            [median - lower_sigma1, upper_sigma1 - median],
            [median - lower_sigma2, upper_sigma2 - median],
        ]



import numpy as np

__all__ = ["Uniform"]


class Uniform(object):
    """Uniform light profile.

    This profile can also compensate for an inaccurate background subtraction.
    name for profile: 'UNIFORM'
    """

    param_names = ["amp"]
    param_names_latex = {r"$I_0$"}
    lower_limit_default = {"amp": -100}
    upper_limit_default = {"amp": 100}

    def __init__(self):
        pass

    def function(self, x, y, amp):
        """

        :param x: x-coordinate
        :param y: y-coordinate
        :param amp: surface brightness
        :return: constant flux
        """
        return np.ones_like(x) * amp



import pytest
import numpy.testing as npt
from lenstronomy.LightModel.Profiles.interpolation import Interpol
from lenstronomy.LightModel.Profiles.gaussian import Gaussian
import lenstronomy.Util.util as util
import numpy as np


class TestInterpol(object):
    """Class to test Shapelets."""

    def setup_method(self):
        pass

    def test_function(self):
        """

        :return:
        """
        for len_x, len_y in [(20, 20), (14, 20)]:
            x, y = util.make_grid(numPix=(len_x, len_y), deltapix=1.0)
            gauss = Gaussian()
            flux = gauss.function(x, y, amp=1.0, center_x=0.0, center_y=0.0, sigma=1.0)
            image = util.array2image(flux, nx=len_y, ny=len_x)

            interp = Interpol()
            kwargs_interp = {
                "image": image,
                "scale": 1.0,
                "phi_G": 0.0,
                "center_x": 0.0,
                "center_y": 0.0,
            }
            output = interp.function(x, y, **kwargs_interp)

            npt.assert_equal(output, flux)

            flux = gauss.function(
                x - 1.0, y, amp=1.0, center_x=0.0, center_y=0.0, sigma=1.0
            )
            kwargs_interp = {
                "image": image,
                "scale": 1.0,
                "phi_G": 0.0,
                "center_x": 1.0,
                "center_y": 0.0,
            }
            output = interp.function(x, y, **kwargs_interp)
            npt.assert_almost_equal(output, flux, decimal=0)

            flux = gauss.function(
                x - 1.0, y - 1.0, amp=1, center_x=0.0, center_y=0.0, sigma=1.0
            )
            kwargs_interp = {
                "image": image,
                "scale": 1.0,
                "phi_G": 0.0,
                "center_x": 1.0,
                "center_y": 1.0,
            }
            output = interp.function(x, y, **kwargs_interp)
            npt.assert_almost_equal(output, flux, decimal=0)

            out = interp.function(x=1000, y=0, **kwargs_interp)
            assert out == 0

        # test change of center without re-doing interpolation
        out = interp.function(
            x=0, y=0, image=image, scale=1.0, phi_G=0, center_x=0, center_y=0
        )
        out_shift = interp.function(
            x=1, y=0, image=image, scale=1.0, phi_G=0, center_x=1, center_y=0
        )
        assert out_shift == out

        # function must give a single value when evaluated at a single point
        assert isinstance(out, float)

        # test change of scale without re-doing interpolation
        out = interp.function(
            x=1.0, y=0, image=image, scale=1.0, phi_G=0, center_x=0, center_y=0
        )
        out_scaled = interp.function(
            x=2.0, y=0, image=image, scale=2, phi_G=0, center_x=0, center_y=0
        )
        assert out_scaled == out

    def test_flux_normalization(self):
        interp = Interpol()
        delta_pix = 0.1
        len_x, len_y = 21, 21
        x, y = util.make_grid(numPix=(len_x, len_y), deltapix=delta_pix)
        gauss = Gaussian()
        flux = gauss.function(x, y, amp=1.0, center_x=0.0, center_y=0.0, sigma=0.3)
        image = util.array2image(flux, nx=len_y, ny=len_x)
        flux_total = np.sum(image)

        kwargs_interp = {
            "image": image,
            "scale": delta_pix,
            "phi_G": 0.0,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        image_interp = interp.function(x, y, **kwargs_interp)
        flux_interp = np.sum(image_interp)
        npt.assert_almost_equal(flux_interp, flux_total, decimal=3)

        # test for scale !=1
        # demands same surface brightness values. We rescale the pixel grid by the same amount as the image
        scale = 0.5
        x, y = util.make_grid(numPix=(len_x, len_y), deltapix=delta_pix * scale)
        kwargs_interp = {
            "image": image,
            "scale": delta_pix * scale,
            "phi_G": 0.0,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        output = interp.function(x, y, **kwargs_interp)

        npt.assert_almost_equal(output / image_interp, 1, decimal=5)

    def test_delete_cache(self):
        x, y = util.make_grid(numPix=20, deltapix=1.0)
        gauss = Gaussian()
        flux = gauss.function(x, y, amp=1.0, center_x=0.0, center_y=0.0, sigma=1.0)
        image = util.array2image(flux)
        interp = Interpol()
        kwargs_interp = {
            "image": image,
            "scale": 1.0,
            "phi_G": 0.0,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        output = interp.function(x, y, **kwargs_interp)
        assert hasattr(interp, "_image_interp")
        interp.delete_cache()
        assert not hasattr(interp, "_image_interp")


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy.testing as npt
import pytest
from lenstronomy.LensModel.Profiles.curved_arc_sis_mst import CurvedArcSISMST
from lenstronomy.LensModel.Profiles.sis import SIS
from lenstronomy.LensModel.Profiles.convergence import Convergence
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions


class TestCurvedArcSISMST(object):
    """Tests the source model routines."""

    def setup_method(self):
        self.model = CurvedArcSISMST()
        self.sis = SIS()
        self.mst = Convergence()

    def test_spp2stretch(self):
        center_x, center_y = 1, 1
        theta_E = 1
        kappa = 0.1
        center_x_spp, center_y_spp = 0.0, 0

        (
            tangential_stretch,
            radial_stretch,
            curvature,
            direction,
        ) = self.model.sis_mst2stretch(
            theta_E, kappa, center_x_spp, center_y_spp, center_x, center_y
        )
        (
            theta_E_new,
            kappa_new,
            center_x_spp_new,
            center_y_spp_new,
        ) = self.model.stretch2sis_mst(
            tangential_stretch, radial_stretch, curvature, direction, center_x, center_y
        )
        npt.assert_almost_equal(center_x_spp_new, center_x_spp, decimal=8)
        npt.assert_almost_equal(center_y_spp_new, center_y_spp, decimal=8)
        npt.assert_almost_equal(theta_E_new, theta_E, decimal=8)
        npt.assert_almost_equal(kappa_new, kappa, decimal=8)

        center_x, center_y = -1, 1
        (
            tangential_stretch,
            radial_stretch,
            curvature,
            direction,
        ) = self.model.sis_mst2stretch(
            theta_E, kappa, center_x_spp, center_y_spp, center_x, center_y
        )
        (
            theta_E_new,
            kappa_new,
            center_x_spp_new,
            center_y_spp_new,
        ) = self.model.stretch2sis_mst(
            tangential_stretch, radial_stretch, curvature, direction, center_x, center_y
        )
        npt.assert_almost_equal(center_x_spp_new, center_x_spp, decimal=8)
        npt.assert_almost_equal(center_y_spp_new, center_y_spp, decimal=8)
        npt.assert_almost_equal(theta_E_new, theta_E, decimal=8)
        npt.assert_almost_equal(kappa_new, kappa, decimal=8)

        center_x, center_y = 0, 0.5
        (
            tangential_stretch,
            radial_stretch,
            curvature,
            direction,
        ) = self.model.sis_mst2stretch(
            theta_E, kappa, center_x_spp, center_y_spp, center_x, center_y
        )
        (
            theta_E_new,
            kappa_new,
            center_x_spp_new,
            center_y_spp_new,
        ) = self.model.stretch2sis_mst(
            tangential_stretch, radial_stretch, curvature, direction, center_x, center_y
        )
        npt.assert_almost_equal(center_x_spp_new, center_x_spp, decimal=8)
        npt.assert_almost_equal(center_y_spp_new, center_y_spp, decimal=8)
        npt.assert_almost_equal(theta_E_new, theta_E, decimal=8)
        npt.assert_almost_equal(kappa_new, kappa, decimal=8)

        center_x, center_y = 0, -1.5
        (
            tangential_stretch,
            radial_stretch,
            r_curvature,
            direction,
        ) = self.model.sis_mst2stretch(
            theta_E, kappa, center_x_spp, center_y_spp, center_x, center_y
        )
        print(tangential_stretch, radial_stretch, r_curvature, direction)
        (
            theta_E_new,
            kappa_new,
            center_x_spp_new,
            center_y_spp_new,
        ) = self.model.stretch2sis_mst(
            tangential_stretch,
            radial_stretch,
            r_curvature,
            direction,
            center_x,
            center_y,
        )
        npt.assert_almost_equal(center_x_spp_new, center_x_spp, decimal=8)
        npt.assert_almost_equal(center_y_spp_new, center_y_spp, decimal=8)
        npt.assert_almost_equal(theta_E_new, theta_E, decimal=8)
        npt.assert_almost_equal(kappa_new, kappa, decimal=8)

    def test_function(self):
        center_x, center_y = 0.0, 0.0
        x, y = 1, 1
        radial_stretch = 1
        output = self.model.function(
            x,
            y,
            tangential_stretch=2,
            radial_stretch=radial_stretch,
            curvature=1.0 / 2,
            direction=0,
            center_x=center_x,
            center_y=center_y,
        )
        theta_E, kappa_ext, center_x_sis, center_y_sis = self.model.stretch2sis_mst(
            tangential_stretch=2,
            radial_stretch=radial_stretch,
            curvature=1.0 / 2,
            direction=0,
            center_x=center_x,
            center_y=center_y,
        )
        f_sis_out = self.sis.function(
            1, 1, theta_E, center_x_sis, center_y_sis
        )  # - self.sis.function(0, 0, theta_E, center_x_sis, center_y_sis)
        alpha_x, alpha_y = self.sis.derivatives(
            center_x, center_y, theta_E, center_x_sis, center_y_sis
        )
        f_sis_0_out = alpha_x * (x - center_x) + alpha_y * (y - center_y)

        f_mst_out = self.mst.function(x, y, kappa_ext, ra_0=center_x, dec_0=center_y)
        lambda_mst = 1.0 / radial_stretch
        f_out = lambda_mst * (f_sis_out - f_sis_0_out) + f_mst_out
        npt.assert_almost_equal(output, f_out, decimal=8)

    def test_derivatives(self):
        tangential_stretch = 5
        radial_stretch = 1
        curvature = 1.0 / 10
        direction = 0.3
        center_x = 0
        center_y = 0
        x, y = 1, 1
        theta_E, kappa, center_x_spp, center_y_spp = self.model.stretch2sis_mst(
            tangential_stretch, radial_stretch, curvature, direction, center_x, center_y
        )
        f_x_sis, f_y_sis = self.sis.derivatives(
            x, y, theta_E, center_x_spp, center_y_spp
        )
        f_x_mst, f_y_mst = self.mst.derivatives(
            x, y, kappa, ra_0=center_x, dec_0=center_y
        )
        f_x0, f_y0 = self.sis.derivatives(
            center_x, center_y, theta_E, center_x_spp, center_y_spp
        )
        f_x_new, f_y_new = self.model.derivatives(
            x,
            y,
            tangential_stretch,
            radial_stretch,
            curvature,
            direction,
            center_x,
            center_y,
        )
        npt.assert_almost_equal(f_x_new, f_x_sis + f_x_mst - f_x0, decimal=8)
        npt.assert_almost_equal(f_y_new, f_y_sis + f_y_mst - f_y0, decimal=8)

    def test_hessian(self):
        lens = LensModel(lens_model_list=["CURVED_ARC_SIS_MST"])
        center_x, center_y = 0, 0
        tangential_stretch = 10
        radial_stretch = 1
        kwargs_lens = [
            {
                "tangential_stretch": tangential_stretch,
                "radial_stretch": radial_stretch,
                "curvature": 1.0 / 10.5,
                "direction": 0.0,
                "center_x": center_x,
                "center_y": center_y,
            }
        ]
        mag = lens.magnification(center_x, center_y, kwargs=kwargs_lens)
        npt.assert_almost_equal(mag, tangential_stretch * radial_stretch, decimal=8)

        center_x, center_y = 2, 3
        tangential_stretch = 10
        radial_stretch = 1
        kwargs_lens = [
            {
                "tangential_stretch": tangential_stretch,
                "radial_stretch": radial_stretch,
                "curvature": 1.0 / 10.5,
                "direction": 0.0,
                "center_x": center_x,
                "center_y": center_y,
            }
        ]
        mag = lens.magnification(center_x, center_y, kwargs=kwargs_lens)
        npt.assert_almost_equal(mag, tangential_stretch * radial_stretch, decimal=8)

        center_x, center_y = 0, 0
        tangential_stretch = 5
        radial_stretch = 1.2
        kwargs_lens = [
            {
                "tangential_stretch": tangential_stretch,
                "radial_stretch": radial_stretch,
                "curvature": 1.0 / 10.5,
                "direction": 0.0,
                "center_x": center_x,
                "center_y": center_y,
            }
        ]
        mag = lens.magnification(center_x, center_y, kwargs=kwargs_lens)
        npt.assert_almost_equal(mag, tangential_stretch * radial_stretch, decimal=8)

        center_x, center_y = 0, 0
        tangential_stretch = 3
        radial_stretch = -1
        kwargs_lens = [
            {
                "tangential_stretch": tangential_stretch,
                "radial_stretch": radial_stretch,
                "curvature": 1.0 / 10.5,
                "direction": 0.0,
                "center_x": center_x,
                "center_y": center_y,
            }
        ]
        mag = lens.magnification(center_x, center_y, kwargs=kwargs_lens)
        print(tangential_stretch, radial_stretch, "stretches")
        npt.assert_almost_equal(mag, tangential_stretch * radial_stretch, decimal=8)

        center_x, center_y = 0, 0
        tangential_stretch = -3
        radial_stretch = -1
        kwargs_lens = [
            {
                "tangential_stretch": tangential_stretch,
                "radial_stretch": radial_stretch,
                "curvature": 1.0 / 10.5,
                "direction": 0.0,
                "center_x": center_x,
                "center_y": center_y,
            }
        ]
        mag = lens.magnification(center_x, center_y, kwargs=kwargs_lens)
        npt.assert_almost_equal(mag, tangential_stretch * radial_stretch, decimal=8)

        center_x, center_y = 0, 0
        tangential_stretch = 10.4
        radial_stretch = 0.6
        kwargs_lens = [
            {
                "tangential_stretch": tangential_stretch,
                "radial_stretch": radial_stretch,
                "curvature": 1.0 / 10.5,
                "direction": 0.0,
                "center_x": center_x,
                "center_y": center_y,
            }
        ]
        mag = lens.magnification(center_x, center_y, kwargs=kwargs_lens)
        npt.assert_almost_equal(mag, tangential_stretch * radial_stretch, decimal=8)

    def test_curved_arc_recovery(self):
        """Test whether the curved arc parameters are satisfied in differential form."""

        ext = LensModelExtensions(LensModel(lens_model_list=["CURVED_ARC_SIS_MST"]))
        center_x, center_y = (
            1,
            1.0,
        )  # test works except at (0,0) where the direction angle is not well defined
        tangential_stretch = 10.0
        radial_stretch = 1.2
        curvature, direction = 0.02, 0.5
        kwargs_lens = {
            "tangential_stretch": tangential_stretch,
            "radial_stretch": radial_stretch,
            "curvature": curvature,
            "direction": direction,
            "center_x": center_x,
            "center_y": center_y,
        }

        self._test_curved_arc_recovery(kwargs_lens)

    def _test_curved_arc_recovery(self, kwargs_arc_init):
        ext = LensModelExtensions(LensModel(lens_model_list=["CURVED_ARC_SIS_MST"]))
        center_x, center_y = kwargs_arc_init["center_x"], kwargs_arc_init["center_y"]
        kwargs_arc = ext.curved_arc_estimate(center_x, center_y, [kwargs_arc_init])
        (
            lambda_rad,
            lambda_tan,
            orientation_angle,
            dlambda_tan_dtan,
            dlambda_tan_drad,
            dlambda_rad_drad,
            dlambda_rad_dtan,
            dphi_tan_dtan,
            dphi_tan_drad,
            dphi_rad_drad,
            dphi_rad_dtan,
        ) = ext.radial_tangential_differentials(center_x, center_y, [kwargs_arc_init])
        npt.assert_almost_equal(
            kwargs_arc["tangential_stretch"],
            kwargs_arc_init["tangential_stretch"],
            decimal=3,
        )
        npt.assert_almost_equal(
            kwargs_arc["radial_stretch"], kwargs_arc_init["radial_stretch"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_arc["curvature"], kwargs_arc_init["curvature"], decimal=3
        )
        npt.assert_almost_equal(dphi_tan_dtan, kwargs_arc_init["curvature"], decimal=3)
        npt.assert_almost_equal(
            kwargs_arc["direction"], kwargs_arc_init["direction"], decimal=3
        )
        npt.assert_almost_equal(dlambda_tan_dtan, 0, decimal=3)


if __name__ == "__main__":
    pytest.main("-k TestLensModel")



__author__ = "ewoudwempe", "sibirrer"

import numpy as np
from lenstronomy.LensModel.Util.epl_util import (
    min_approx,
    pol_to_cart,
    cart_to_pol,
    cdot,
    ps,
    rotmat,
    solvequadeq,
    brentq_inline,
)
from lenstronomy.LensModel.Util.epl_util import pol_to_ell, ell_to_pol, geomlinspace
from lenstronomy.Util.image_util import findOverlap
from lenstronomy.LensModel.Profiles.epl_numba import alpha, omega
from lenstronomy.Util.numba_util import jit
from lenstronomy.Util.param_util import (
    ellipticity2phi_q,
    shear_cartesian2polar,
    shear_polar2cartesian,
)
from lenstronomy.LensModel.Profiles.shear import Shear


@jit()
def _alpha_epl_shear(x, y, b, q, t=1, gamma1=0, gamma2=0, Omega=None):
    """The complex deflection of EPL+SHEAR."""
    return (
        alpha(x, y, b, q, t=t, Omega=Omega)
        + (gamma1 * x + gamma2 * y)
        + 1j * (gamma2 * x - gamma1 * y)
    )


@jit()
def _one_dim_lens_eq_calcs(args, phi):
    """Calculates intermediate quantities that are needed for several of the subsequent
    functions."""
    b, t, y1, y2, q, gamma1, gamma2 = args
    y = y1 + y2 * 1j

    rhat = (
        1
        / (1 - gamma1**2 - gamma2**2)
        * (
            ((1 + gamma1) * np.cos(phi) + gamma2 * np.sin(phi))
            + 1j * (gamma2 * np.cos(phi) + (1 - gamma1) * np.sin(phi))
        )
    )
    thetahat = (
        1
        / (1 - gamma1**2 - gamma2**2)
        * (
            ((1 + gamma1) * np.sin(phi) - gamma2 * np.cos(phi))
            + 1j * (gamma2 * np.sin(phi) - (1 - gamma1) * np.cos(phi))
        )
    )

    frac_Roverrsh, phiell = pol_to_ell(1, phi, q)
    Omega = omega(phiell, t, q)
    const = 2 * b / (1 + q)
    if abs(t - 1) > 1e-4:
        b_over_r_pow_tm1 = -cdot(y, thetahat) / (const * cdot(Omega, thetahat))
        R = b * np.abs(b_over_r_pow_tm1) ** (1 / (1 - t)) * np.sign(b_over_r_pow_tm1)
    else:
        Omega_ort = 1j * Omega
        x = ((1 - gamma1) * np.cos(phi) - gamma2 * np.sin(phi)) + 1j * (
            -gamma2 * np.cos(phi) + (1 + gamma1) * np.sin(phi)
        )
        R = cdot(Omega_ort, y) / cdot(Omega_ort, x) * frac_Roverrsh
    r, theta = ell_to_pol(R, phiell, q)
    return Omega, const, phiell, q, r, rhat, t, b, thetahat, y


@jit()
def _one_dim_lens_eq_both(phi, args):
    """Calculates and returns simultaneously both the smooth and the not-smooth
    1-dimensional lens equation that needs to be solved."""
    Omega, const, phiell, q, r, rhat, t, b, thetahat, y = _one_dim_lens_eq_calcs(
        args, phi
    )
    rr, thetaa = ell_to_pol(1, phiell, q)
    ip = cdot(y, rhat) * cdot(Omega, thetahat) - cdot(Omega, rhat) * cdot(y, thetahat)
    # The derivations are lost somewhere in my notes...
    eq = (rr * b) ** (2 / t - 2) * ps(
        (cdot(y, thetahat) / const), 2 / t
    ) * ip**2 + ps(ip, 2 / t) * cdot(Omega, thetahat) ** 2
    eq_notsmooth = ps(rr * b, 1 - t) * (cdot(y, thetahat) / const) * np.abs(
        ip
    ) ** t + ip * np.abs(cdot(Omega, thetahat)) ** (+t)
    return eq, eq_notsmooth


@jit()
def _getr(phi, args):
    """Given an angle phi, get the radius r."""
    Omega, const, phiell, q, r, rhat, t, b, thetahat, y = _one_dim_lens_eq_calcs(
        args, phi
    )
    return r


@jit()
def _one_dim_lens_eq(phi, args):
    """Calculates the smooth 1-dimensional lens equation to solve - to be used by a root-finder"""
    Omega, const, phiell, q, r, rhat, t, b, thetahat, y = _one_dim_lens_eq_calcs(
        args, phi
    )

    rr, thetaa = ell_to_pol(1, phiell, q)
    ip = cdot(y, rhat) * cdot(Omega, thetahat) - cdot(Omega, rhat) * cdot(y, thetahat)
    eq = (rr * b) ** (2 / t - 2) * ps(
        (cdot(y, thetahat) / const), 2 / t
    ) * ip**2 + ps(ip, 2 / t) * cdot(Omega, thetahat) ** 2
    return eq


@jit()
def _one_dim_lens_eq_unsmooth(phi, args):
    """Calculates the not-smooth 1-dimensional lens equation to to solve - to be used by a root-finder.
    For some parameters and solutions, numerical issues make solving this one feasible while the other is not.
    """
    Omega, const, phiell, q, r, rhat, t, b, thetahat, y = _one_dim_lens_eq_calcs(
        args, phi
    )

    rr, thetaa = ell_to_pol(1, phiell, q)
    ip = cdot(y, rhat) * cdot(Omega, thetahat) - cdot(Omega, rhat) * cdot(y, thetahat)
    eq_notsmooth = ps(rr * b, 1 - t) * (cdot(y, thetahat) / const) * np.abs(
        ip
    ) ** t + ip * np.abs(cdot(Omega, thetahat)) ** (+t)
    return eq_notsmooth


@jit()
def _getphi(thpl, args):
    """Finds all roots to both versions the 1-dimensional lens equation in phi, by doing
    a grid search for sign changes on the supplied thpl. In the case of extrema, refine
    at the relevant location.

    :param thpl: What points to calculate the equation on use for detecting sign changes
    :param args: Parameters to be passed to the lens equation
    :return: an array containing all roots
    """
    y, y_notsmooth = _one_dim_lens_eq_both(thpl, args)
    num_phi = len(thpl)
    roots = []
    for i in range(num_phi - 1):
        if y[i + 1] * y[i] <= 0:
            roots.append(
                brentq_inline(_one_dim_lens_eq, thpl[i], thpl[i + 1], args=args)
                % (2 * np.pi)
            )
        elif y_notsmooth[i + 1] * y_notsmooth[i] <= 0:
            roots.append(
                brentq_inline(
                    _one_dim_lens_eq_unsmooth, thpl[i], thpl[i + 1], args=args
                )
                % (2 * np.pi)
            )

    for i in range(1, num_phi + 1):
        y1 = y[i - 1]
        y2 = y[i % num_phi]
        y3 = y[(i + 1) % num_phi]
        y1n = y_notsmooth[(i - 1) % num_phi]
        y2n = y_notsmooth[i % num_phi]
        y3n = y_notsmooth[(i + 1) % num_phi]
        if (y3 - y2) * (y2 - y1) <= 0 or (y3n - y2n) * (y2n - y1n) <= 0:
            if y3 * y2 <= 0 or y1 * y2 <= 0:
                continue
            if i > num_phi - 2:
                continue
            else:
                x1 = thpl[i - 1]
                x2 = thpl[i]
                x3 = thpl[(i + 1)]

            xmin = min_approx(x1, x2, x3, y1, y2, y3)
            xmin_ns = min_approx(x1, x2, x3, y1n, y2n, y3n)
            ymin = _one_dim_lens_eq(xmin, args)
            ymin_ns = _one_dim_lens_eq_unsmooth(xmin_ns, args)
            if ymin * y2 <= 0 and x2 <= xmin <= x3:
                roots.append(
                    brentq_inline(_one_dim_lens_eq, x2, xmin, args=args) % (2 * np.pi)
                )
                roots.append(
                    brentq_inline(_one_dim_lens_eq, xmin, x3, args=args) % (2 * np.pi)
                )
            elif ymin * y2 <= 0 and x1 <= xmin <= x2:
                roots.append(
                    brentq_inline(_one_dim_lens_eq, x1, xmin, args=args) % (2 * np.pi)
                )
                roots.append(
                    brentq_inline(_one_dim_lens_eq, xmin, x2, args=args) % (2 * np.pi)
                )
            elif ymin_ns * y2n <= 0 and x2 <= xmin_ns <= x3:
                roots.append(
                    brentq_inline(_one_dim_lens_eq_unsmooth, x2, xmin_ns, args=args)
                    % (2 * np.pi)
                )
                roots.append(
                    brentq_inline(_one_dim_lens_eq_unsmooth, xmin_ns, x3, args=args)
                    % (2 * np.pi)
                )
            elif ymin_ns * y2n <= 0 and x1 <= xmin_ns <= x2:
                roots.append(
                    brentq_inline(_one_dim_lens_eq_unsmooth, x1, xmin_ns, args=args)
                    % (2 * np.pi)
                )
                roots.append(
                    brentq_inline(_one_dim_lens_eq_unsmooth, xmin_ns, x2, args=args)
                    % (2 * np.pi)
                )

    return np.array(roots)


def solvelenseq_majoraxis(args, Nmeas=200, Nmeas_extra=50):
    """Solve the lens equation, where the arguments have been properly rotated to the
    major-axis."""
    b, t, y1, y2, q, gamma1, gamma2 = args
    p1 = np.arctan2(y2 * (1 - gamma1) + gamma2 * y1, y1 * (1 + gamma1) + gamma2 * y2)
    int_points = [p1]
    geom = geomlinspace(1e-4, 0.1, Nmeas_extra)

    thpl = np.sort(
        np.concatenate(
            (
                np.linspace(0.0, np.pi, Nmeas),
                *[i % np.pi - geom for i in int_points],
                *[i % np.pi + geom for i in int_points],
            )
        )
    )
    the = _getphi(thpl, (b, t, y1, y2, q, gamma1, gamma2))
    thetas = np.concatenate((the, the + np.pi))
    Rs = np.array([_getr(theta, (b, t, y1, y2, q, gamma1, gamma2)) for theta in thetas])
    stuff = np.array(pol_to_cart(Rs[Rs > 0], thetas[Rs > 0]))
    diff = (
        -y1
        - y2 * 1j
        + stuff[0]
        + stuff[1] * 1j
        - _alpha_epl_shear(stuff[0], stuff[1], b, q, t, gamma1=gamma1, gamma2=gamma2)
    )
    goodones = np.abs(diff) < 1e-8
    return findOverlap(*stuff[:, goodones], 1e-8)


def _check_center(kwargs_lens):
    """Checks if the shear-at-center convention is properly used."""
    # calculate (inverse) displacement caused by the offset between shear and lens centroid
    # this shift needs to be added to the source position such that the solution of the lens equation
    # without this shift in the shear is the correct one
    if len(kwargs_lens) > 1:
        shear = Shear()
        # calculate shift from the deflector centroid from the shear field
        alpha_x, alpha_y = shear.derivatives(
            kwargs_lens[0]["center_x"], kwargs_lens[0]["center_y"], **kwargs_lens[1]
        )
        return alpha_x, alpha_y
    else:
        return 0, 0


def solve_lenseq_pemd(pos_, kwargs_lens, Nmeas=400, Nmeas_extra=80, **kwargs):
    """Solves the lens equation using a semi-analytical recipe.

    :param pos_: The source plane position (shape (2,)), or the source plane positions (shape (2,N)) for which to solve the lens equation
    :param kwargs_lens: List of kwargs in lenstronomy style, following ['EPL', 'SHEAR'] format
    :param Nmeas: resolution with which to sample the angular grid, higher means more reliable lens equation solving. For solving many positions at once, you may want to set this higher.
    :param Nmeas_extra: resolution with which to additionally sample the angular grid at the low-shear end, higher means more reliable lens equation solving. For solving many positions at once, you may want to set this higher.
    :return: The lens plane positions.
    Note: generally the (demagnified) central image will also be included.
    """
    pos = np.asarray(pos_)
    if pos.ndim > 1 and pos.shape[-1] != 1:
        pos = pos[..., None]
    t = kwargs_lens[0]["gamma"] - 1 if "gamma" in kwargs_lens[0] else 1

    theta_ell, q = ellipticity2phi_q(kwargs_lens[0]["e1"], kwargs_lens[0]["e2"])
    b = kwargs_lens[0]["theta_E"] * np.sqrt(q)
    if len(kwargs_lens) > 1:
        gamma = kwargs_lens[1]["gamma1"] + 1j * kwargs_lens[1]["gamma2"]
    else:
        gamma = 0 + 0j
    shift_x, shift_y = _check_center(kwargs_lens)
    shift = shift_x + 1j * shift_y
    cen = kwargs_lens[0]["center_x"] + 1j * kwargs_lens[0]["center_y"]
    p = pos[0] + 1j * pos[1] - cen + shift

    rotfact = np.exp(-1j * theta_ell)
    gamma *= rotfact**2
    p *= rotfact
    res = solvelenseq_majoraxis(
        (b, t, p.real, p.imag, q, gamma.real, gamma.imag),
        Nmeas=Nmeas,
        Nmeas_extra=Nmeas_extra,
    )
    xsol, ysol = res
    x = np.array([(xs + 1j * ys) / rotfact + cen for xs, ys in zip(xsol, ysol)])
    return x.real, x.imag


def caustics_epl_shear(
    kwargs_lens, num_th=500, maginf=0, sourceplane=True, return_which=None
):
    """Analytically calculates the caustics of an EPL+shear lens model. Since for
    gamma>2, the outer critical curve does not exist, the option to find the curves for
    a set, finite magnification exists, by supplying maginf, so that the routine finds
    the curve of this magnification, rather than the true caustic.

    :param kwargs_lens: List of kwargs in lenstronomy style, following ['EPL', 'SHEAR']
        format
    :param num_th: resolution.
    :param maginf: the outer critical curve for t>1 will be replaced with the curve
        where the inverse magnification is maginf
    :param sourceplane: if True (default), ray-shoot the calculated critical curves to
        the source plane
    :param return_which: options 'quad' (boundary of area within which there are 4
        images), 'double' (boundary of area within which there are 2 images), 'caustic'
        (the diamond caustic) and 'cut' (the cut, if it exists, that is if t<2, else, if
        t>2, returns the caustic) and None (in that case: return quad, caustic, cut)
    :return: (2,N) array if return_which set, else a tuple of (caustic, cut, quad)
    """
    e1, e2 = kwargs_lens[0]["e1"], kwargs_lens[0]["e2"]
    if len(kwargs_lens) > 1:
        gamma1unr, gamma2unr = kwargs_lens[1]["gamma1"], kwargs_lens[1]["gamma2"]
    else:
        gamma1unr, gamma2unr = 0, 0
    t = kwargs_lens[0]["gamma"] - 1 if "gamma" in kwargs_lens[0] else 1
    theta_ell, q = ellipticity2phi_q(e1, e2)
    theta_gamma, gamma_mag = shear_cartesian2polar(gamma1unr, gamma2unr)
    b = np.sqrt(q) * kwargs_lens[0]["theta_E"]
    cen = np.expand_dims(
        np.array([kwargs_lens[0]["center_x"], kwargs_lens[0]["center_y"]]), 1
    )
    theta_gamma -= theta_ell
    gamma1, gamma2 = shear_polar2cartesian(theta_gamma, gamma_mag)
    M = rotmat(-theta_ell)
    phiell, q = ellipticity2phi_q(e1, e2)
    theta = np.linspace(0, 2 * np.pi, num_th, endpoint=False)
    r = 1
    R, phi = pol_to_ell(1, theta, q)
    Omega = omega(phi, t, q)
    aa = 1
    bb = -(2 - t)
    frac_roverR = r / R
    cc = (
        (1 - t)
        * (2 - t)
        * (cdot(np.exp(1j * theta), Omega))
        / frac_roverR
        * 2
        / (1 + q)
    )
    cc -= (1 - t) ** 2 * (2 / (1 + q)) ** 2 * np.abs(Omega) ** 2 / frac_roverR**2
    # Shear stuff:
    gammaint_fac = (
        -np.exp(2j * theta) * (2 - t) / 2
        + (1 - t) * np.exp(1j * theta) * 2 / (1 + q) * Omega / frac_roverR
    )
    gamma = gamma1 + 1j * gamma2
    aa -= np.abs(gamma) ** 2
    bb -= 2 * cdot(gamma, gammaint_fac)
    usol = np.array(solvequadeq(cc, bb, aa)).T
    xcr_4, ycr_4 = pol_to_cart(b * usol[:, 1] ** (-1 / t) * frac_roverR, theta)
    if (
        t > 1
    ):  # If t>1, get the approximate outer caustic instead (where inverse magnification = maginf).
        usol = np.array(solvequadeq(cc, bb, aa - maginf)).T
        xcr_cut, ycr_cut = pol_to_cart(b * usol[:, 1] ** (-1 / t) * frac_roverR, theta)
    else:
        usol = np.array(solvequadeq(cc, bb, aa + maginf)).T
        xcr_cut, ycr_cut = pol_to_cart(b * usol[:, 0] ** (-1 / t) * frac_roverR, theta)
    al_cut = _alpha_epl_shear(xcr_cut, ycr_cut, b, q, t, gamma1, gamma2, Omega=Omega)
    al_4 = _alpha_epl_shear(xcr_4, ycr_4, b, q, t, gamma1, gamma2, Omega=Omega)
    if sourceplane:
        xca_cut, yca_cut = xcr_cut - al_cut.real, ycr_cut - al_cut.imag
        xca_4, yca_4 = xcr_4 - al_4.real, ycr_4 - al_4.imag
    else:
        xca_cut, yca_cut = xcr_cut, ycr_cut
        xca_4, yca_4 = xcr_4, ycr_4
    if return_which == "caustic":
        return M @ (xca_4, yca_4) + cen
    if return_which == "cut":
        return M @ (xca_cut, yca_cut) + cen

    rcut, thcut = cart_to_pol(xca_cut, yca_cut)
    r, th = cart_to_pol(xca_4, yca_4)
    r2 = np.interp(th, thcut, rcut, period=2 * np.pi)

    if return_which == "double":
        r = np.fmax(r, r2)
    else:  # Quad
        r = np.fmin(r, r2)

    pos_tosample = np.empty((2, num_th))
    pos_tosample[0], pos_tosample[1] = pol_to_cart(r, th)
    if return_which in ("double", "quad"):
        return M @ pos_tosample + cen

    return (
        M @ (xca_4, yca_4) + cen,
        M @ (xca_cut, yca_cut) + cen,
        M @ pos_tosample + cen,
    )  # Mostly for some backward compatibility



__author__ = "sibirrer"

import numpy as np
import lenstronomy.Util.util as util
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["NIE", "NIEMajorAxis"]


class NIE(LensProfileBase):
    """Non-singular isothermal ellipsoid (NIE)

    .. math::
     \\kappa = \\theta_E/2 \\left[s^2_{scale} + qx^2 + y^2/q]−1/2
    """

    param_names = ["theta_E", "e1", "e2", "s_scale", "center_x", "center_y"]
    lower_limit_default = {
        "theta_E": 0,
        "e1": -0.5,
        "e2": -0.5,
        "s_scale": 0,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "theta_E": 10,
        "e1": 0.5,
        "e2": 0.5,
        "s_scale": 100,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.nie_major_axis = NIEMajorAxis()
        super(NIE, self).__init__()

    def function(self, x, y, theta_E, e1, e2, s_scale, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale
        :param center_x: profile center
        :param center_y: profile center
        :return: lensing potential
        """
        b, s, q, phi_G = self.param_conv(theta_E, e1, e2, s_scale)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)
        # evaluate
        f_ = self.nie_major_axis.function(x__, y__, b, s, q)
        # rotate back
        return f_

    def derivatives(self, x, y, theta_E, e1, e2, s_scale, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale
        :param center_x: profile center
        :param center_y: profile center
        :return: alpha_x, alpha_y
        """
        b, s, q, phi_G = self.param_conv(theta_E, e1, e2, s_scale)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)
        # evaluate
        f__x, f__y = self.nie_major_axis.derivatives(x__, y__, b, s, q)
        # rotate back
        f_x, f_y = util.rotate(f__x, f__y, -phi_G)
        return f_x, f_y

    def hessian(self, x, y, theta_E, e1, e2, s_scale, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale
        :param center_x: profile center
        :param center_y: profile center
        :return: f_xx, f_xy, f_yx, f_yy
        """
        b, s, q, phi_G = self.param_conv(theta_E, e1, e2, s_scale)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)
        # evaluate
        f__xx, f__xy, _, f__yy = self.nie_major_axis.hessian(x__, y__, b, s, q)
        # rotate back
        kappa = 1.0 / 2 * (f__xx + f__yy)
        gamma1__ = 1.0 / 2 * (f__xx - f__yy)
        gamma2__ = f__xy
        gamma1 = np.cos(2 * phi_G) * gamma1__ - np.sin(2 * phi_G) * gamma2__
        gamma2 = +np.sin(2 * phi_G) * gamma1__ + np.cos(2 * phi_G) * gamma2__
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_xy, f_xy, f_yy

    def density_lens(self, r, theta_E, e1, e2, s_scale, center_x=0, center_y=0):
        """3d mass density at 3d radius r. This function assumes spherical
        symmetry/ignoring the eccentricity.

        :param r: 3d radius
        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale
        :param center_x: profile center
        :param center_y: profile center
        :return: 3d mass density at 3d radius r
        """
        # kappa=1/2 at Einstein radius
        rho0 = 1 / 2 * theta_E / np.pi
        return rho0 / (r**2 + s_scale**2)

    def mass_3d_lens(self, r, theta_E, e1, e2, s_scale, center_x=0, center_y=0):
        """Mass enclosed a 3d radius r. This function assumes spherical
        symmetry/ignoring the eccentricity.

        :param r: 3d radius
        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale
        :param center_x: profile center
        :param center_y: profile center
        :return: 3d mass density at 3d radius r
        """
        rho0 = 1 / 2 * theta_E / np.pi
        return rho0 * 4 * np.pi * (r - s_scale * np.arctan(r / s_scale))

    def param_conv(self, theta_E, e1, e2, s_scale):
        if self._static is True:
            return self._b_static, self._s_static, self._q_static, self._phi_G_static
        return self._param_conv(theta_E, e1, e2, s_scale)

    def _param_conv(self, theta_E, e1, e2, s_scale):
        """
        convert parameters from 2*kappa = bIE [s2IE + r2(1 − e *cos(2*phi)]−1/2 to
        2*kappa=  b *(q2(s2 + x2) + y2􏰉)−1/2
        see expressions after Equation 8 in Keeton and Kochanek 1998, https://arxiv.org/pdf/astro-ph/9705194.pdf

        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale
        :return: critical radius b, smoothing scale s, axis ratio q, orientation angle phi_G
        """

        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        theta_E_conv = self._theta_E_prod_average2major_axis(theta_E, q)
        b = theta_E_conv * np.sqrt((1 + q**2) / 2)
        s = s_scale / np.sqrt(q)
        # s = s_scale * np.sqrt((1 + q**2) / (2*q**2))
        return b, s, q, phi_G

    def set_static(self, theta_E, e1, e2, s_scale, center_x=0, center_y=0):
        """

        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale
        :param center_x: profile center
        :param center_y: profile center
        :return: self variables set
        """
        self._static = True
        (
            self._b_static,
            self._s_static,
            self._q_static,
            self._phi_G_static,
        ) = self._param_conv(theta_E, e1, e2, s_scale)

    def set_dynamic(self):
        """

        :return:
        """
        self._static = False
        if hasattr(self, "_b_static"):
            del self._b_static
        if hasattr(self, "_s_static"):
            del self._s_static
        if hasattr(self, "_phi_G_static"):
            del self._phi_G_static
        if hasattr(self, "_q_static"):
            del self._q_static

    @staticmethod
    def _theta_E_prod_average2major_axis(theta_E, q):
        """Converts a product averaged Einstein radius (of semi-minor and semi-major
        axis) to a major axis Einstein radius for an Isothermal ellipse. The standard
        lenstronomy conventions are product averaged Einstein radii while other codes
        (such as e.g. gravlens) use the semi-major axis convention.

        .. math::
          \\frac{\\theta_{E, prod ave}}{\\theta_{E, major}} = \\sqrt{(1+q^2) / (2 q) }

        :param theta_E: Einstein radius in lenstronomy conventions (product average of major and minor axes)
        :param q: axis ratio minor/major
        :return: theta_E in convention of kappa= b *(q^2(s^2 + x^2) + y^2􏰉)^{−1/2} (major axis)
        """
        theta_E_major_axis = theta_E / (np.sqrt((1.0 + q**2) / (2.0 * q)))
        return theta_E_major_axis


class NIEMajorAxis(LensProfileBase):
    """
    This class contains the function and the derivatives of the non-singular isothermal ellipse.
    See Keeton and Kochanek 1998, https://arxiv.org/pdf/astro-ph/9705194.pdf

    .. math::
        \\kappa =  b * (q2(s2 + x2) + y2􏰉)^{−1/2}`

    """

    param_names = ["b", "s", "q", "center_x", "center_y"]

    def __init__(self, diff=0.0000000001):
        self._diff = diff
        super(NIEMajorAxis, self).__init__()

    def function(self, x, y, b, s, q):
        psi = self._psi(x, y, q, s)
        alpha_x, alpha_y = self.derivatives(x, y, b, s, q)
        f_ = (
            x * alpha_x
            + y * alpha_y
            - b * s * 1.0 / 2.0 * np.log((psi + s) ** 2 + (1.0 - q**2) * x**2)
        )
        return f_

    def derivatives(self, x, y, b, s, q):
        """Returns df/dx and df/dy of the function."""
        if q >= 1:
            q = 0.99999999
        psi = self._psi(x, y, q, s)
        f_x = (
            b / np.sqrt(1.0 - q**2) * np.arctan(np.sqrt(1.0 - q**2) * x / (psi + s))
        )
        f_y = (
            b
            / np.sqrt(1.0 - q**2)
            * np.arctanh(np.sqrt(1.0 - q**2) * y / (psi + q**2 * s))
        )
        return f_x, f_y

    def hessian(self, x, y, b, s, q):
        """Returns Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx,
        d^f/dy^2."""
        alpha_ra, alpha_dec = self.derivatives(x, y, b, s, q)
        diff = self._diff
        alpha_ra_dx, alpha_dec_dx = self.derivatives(x + diff, y, b, s, q)
        alpha_ra_dy, alpha_dec_dy = self.derivatives(x, y + diff, b, s, q)

        f_xx = (alpha_ra_dx - alpha_ra) / diff
        f_xy = (alpha_ra_dy - alpha_ra) / diff
        f_yx = (alpha_dec_dx - alpha_dec) / diff
        f_yy = (alpha_dec_dy - alpha_dec) / diff
        return f_xx, f_xy, f_yx, f_yy

    @staticmethod
    def kappa(x, y, b, s, q):
        """convergence.

        :param x: major axis coordinate
        :param y: minor axis coordinate
        :param b: normalization
        :param s: smoothing scale
        :param q: axis ratio
        :return: convergence
        """
        kappa = b / 2.0 * (q**2 * (s**2 + x**2) + y**2) ** (-1.0 / 2)
        return kappa

    @staticmethod
    def _psi(x, y, q, s):
        """Expression after equation (8) in Keeton&Kochanek 1998.

        :param x: semi-major axis coordinate
        :param y: semi-minor axis coordinate
        :param q: axis ratio minor/major
        :param s: smoothing scale in major axis direction
        :return: phi
        """
        return np.sqrt(q**2 * (s**2 + x**2) + y**2)



__author__ = "aymgal"

import os
import json
import shutil
import numpy as np

from lenstronomy.Sampling.Samplers.base_nested_sampler import NestedSampler
import lenstronomy.Util.sampling_util as utils

__all__ = ["MultiNestSampler"]


class MultiNestSampler(NestedSampler):
    """Wrapper for nested sampling algorithm MultInest by F.

    Feroz & M. Hobson
    papers : arXiv:0704.3704, arXiv:0809.3437, arXiv:1306.2144
    pymultinest doc : https://johannesbuchner.github.io/PyMultiNest/pymultinest.html
    """

    def __init__(
        self,
        likelihood_module,
        prior_type="uniform",
        prior_means=None,
        prior_sigmas=None,
        width_scale=1,
        sigma_scale=1,
        output_dir=None,
        output_basename="-",
        remove_output_dir=False,
        use_mpi=False,
    ):
        """
        :param likelihood_module: likelihood_module like in likelihood.py (should be callable)
        :param prior_type: 'uniform' of 'gaussian', for converting the unit hypercube to param cube
        :param prior_means: if prior_type is 'gaussian', mean for each param
        :param prior_sigmas: if prior_type is 'gaussian', std dev for each param
        :param width_scale: scale the widths of the parameters space by this factor
        :param sigma_scale: if prior_type is 'gaussian', scale the gaussian sigma by this factor
        :param output_dir: name of the folder that will contain output files
        :param output_basename: prefix for output files
        :param remove_output_dir: remove the output_dir folder after completion
        :param use_mpi: flag directly passed to MultInest sampler (NOT TESTED)
        """
        self._check_install()
        super(MultiNestSampler, self).__init__(
            likelihood_module,
            prior_type,
            prior_means,
            prior_sigmas,
            width_scale,
            sigma_scale,
        )

        # here we assume number of dimensons = number of parameters
        self.n_params = self.n_dims

        if output_dir is None:
            self._output_dir = "multinest_out_default"
        else:
            self._output_dir = output_dir

        self._is_master = True
        self._use_mpi = use_mpi

        if self._use_mpi:
            from mpi4py import MPI

            self._comm = MPI.COMM_WORLD

            if self._comm.Get_rank() != 0:
                self._is_master = False
        else:
            self._comm = None

        if self._is_master:
            if os.path.exists(self._output_dir):
                shutil.rmtree(self._output_dir, ignore_errors=True)
            os.mkdir(self._output_dir)

        self.files_basename = os.path.join(self._output_dir, output_basename)

        # required for analysis : save parameter names in json file
        if self._is_master:
            with open(self.files_basename + "params.json", "w") as file:
                json.dump(self.param_names, file, indent=2)

        self._rm_output = remove_output_dir

    def run(self, kwargs_run):
        """Run the MultiNest nested sampler.

        see https://johannesbuchner.github.io/PyMultiNest/pymultinest.html for content
        of kwargs_run

        :param kwargs_run: kwargs directly passed to pymultinest.run
        :return: samples, means, logZ, logZ_err, logL, stats
        """
        print("prior type :", self.prior_type)
        print("parameter names :", self.param_names)

        if self._pymultinest_installed:
            self._pymultinest.run(
                self.log_likelihood,
                self.prior,
                self.n_dims,
                outputfiles_basename=self.files_basename,
                resume=False,
                verbose=True,
                init_MPI=self._use_mpi,
                **kwargs_run
            )

            analyzer = self._Analyzer(
                self.n_dims, outputfiles_basename=self.files_basename
            )
            samples = analyzer.get_equal_weighted_posterior()[:, :-1]
            data = analyzer.get_data()  # gets data from the *.txt output file
            stats = analyzer.get_stats()

        else:
            # in case MultiNest was not compiled properly, for unit tests
            samples = np.zeros((1, self.n_dims))
            data = np.zeros((self.n_dims, 3))
            stats = {
                "global evidence": np.zeros(self.n_dims),
                "global evidence error": np.zeros(self.n_dims),
                "modes": [{"mean": np.zeros(self.n_dims)}],
            }

        logL = -0.5 * data[:, 1]  # since the second data column is -2*logL
        logZ = stats["global evidence"]
        logZ_err = stats["global evidence error"]
        # or better to use stats['marginals'][:]['median'] ???
        means = stats["modes"][0]["mean"]

        print(
            "MultiNest output files have been saved to {}*".format(self.files_basename)
        )

        if self._rm_output and self._is_master:
            shutil.rmtree(self._output_dir, ignore_errors=True)
            print("MultiNest output directory removed")

        return samples, means, logZ, logZ_err, logL, stats

    def _check_install(self):
        try:
            import pymultinest
            from pymultinest.analyse import Analyzer
        except:
            print(
                "Warning : MultiNest/pymultinest not properly installed (results might be unexpected). \
                    You can get it from : https://johannesbuchner.github.io/PyMultiNest/pymultinest.html"
            )
            self._pymultinest_installed = False
        else:
            self._pymultinest_installed = True
            self._pymultinest = pymultinest
            self._Analyzer = Analyzer



import numpy as np
import lenstronomy.Util.util as util
from lenstronomy.Util import param_util

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
def mask_center_2d(center_x, center_y, r, x_grid, y_grid):
    """

    :param center_x: x-coordinate of center position of circular mask
    :param center_y: y-coordinate of center position of circular mask
    :param r: radius of mask in pixel values
    :param x_grid: x-coordinate grid
    :param y_grid: y-coordinate grid
    :return: mask array of shape x_grid with =0 inside the radius and =1 outside
    :rtype: array of size of input grid with integers 0 or 1
    """
    x_shift = x_grid - center_x
    y_shift = y_grid - center_y
    R = np.sqrt(x_shift * x_shift + y_shift * y_shift)
    mask = np.empty_like(R, dtype="int")
    mask[R > r] = 1
    mask[R <= r] = 0
    return mask


@export
def mask_azimuthal(x, y, center_x, center_y, r):
    """Azimuthal mask with =1 inside radius and =0 outside.

    :param x: x-coordinates (1d or 2d array numpy array)
    :param y: y-coordinates (1d or 2d array numpy array)
    :param center_x: center of azimuthal mask in x
    :param center_y: center of azimuthal mask in y
    :param r: radius of azimuthal mask
    :return: array with zeros outside r and ones inside azimuthal radius r
    :rtype: array of size of input grid with integers 0 or 1
    """
    x_shift = x - center_x
    y_shift = y - center_y
    R = np.sqrt(x_shift * x_shift + y_shift * y_shift)
    mask = np.empty_like(R, dtype="int")
    mask[R > r] = 0
    mask[R <= r] = 1
    return mask


@export
def mask_ellipse(x, y, center_x, center_y, a, b, angle):
    """

    :param x: x-coordinates of pixels
    :param y: y-coordinates of pixels
    :param center_x: center of mask
    :param center_y: center of mask
    :param a: major axis
    :param b: minor axis
    :param angle: angle of major axis
    :return: mask (list of zeros and ones)
    :rtype: array of size of input grid with integers 0 or 1
    """
    x_shift = x - center_x
    y_shift = y - center_y
    x_rot, y_rot = util.rotate(x_shift, y_shift, angle)
    r_ab = x_rot**2 / a**2 + y_rot**2 / b**2
    mask = np.empty_like(r_ab, dtype="int")
    mask[r_ab > 1] = 0
    mask[r_ab <= 1] = 1
    return mask


@export
def mask_eccentric(x, y, center_x, center_y, e1, e2, r):
    """Elliptical mask with eccentricities as input.

    :param x: x-coordinate array
    :param y: y-coordinate array
    :param center_x: center in x
    :param center_y: center in y
    :param e1: eccentricity e1
    :param e2: eccentricity e2
    :param r: radius of mask
    :return: mask (list of zeros and ones)
    :rtype: array of size of input grid with integers 0 or 1
    """
    x_, y_ = param_util.transform_e1e2_product_average(x, y, e1, e2, center_x, center_y)
    r_ab = (x_**2 + y_**2) / r**2
    mask = np.empty_like(r_ab, dtype="int")
    mask[r_ab > 1] = 0
    mask[r_ab <= 1] = 1
    return mask


@export
def mask_half_moon(x, y, center_x, center_y, r_in, r_out, phi0=0, delta_phi=2 * np.pi):
    """

    :param x:
    :param y:
    :param center_x:
    :param center_y:
    :param r_in:
    :param r_out:
    :param phi0:
    :param delta_phi:
    :return:
    :rtype: array of size of input grid with integers 0 or 1
    """
    x_shift = x - center_x
    y_shift = y - center_y
    R = np.sqrt(x_shift * x_shift + y_shift * y_shift)
    phi = np.arctan2(x_shift, y_shift)
    phi_min = phi0 - delta_phi / 2.0
    phi_max = phi0 + delta_phi / 2.0
    mask = np.zeros_like(x, dtype="int")
    if phi_max > phi_min:
        mask[(R < r_out) & (R > r_in) & (phi > phi_min) & (phi < phi_max)] = 1
    else:
        mask[(R < r_out) & (R > r_in) & (phi > phi_max)] = 1
        mask[(R < r_out) & (R > r_in) & (phi < phi_min)] = 1
    return mask


def mask_shell(x, y, center_x, center_y, r_in, r_out):
    """Ring mask.

    :param x: x-coordinate grid
    :param y: y-coordinate grid
    :param center_x: center of ring mask
    :param center_y: center of ring mask
    :param r_in: inner ring radius
    :param r_out: outer ring radius
    :return: mask
    """
    x_shift = x - center_x
    y_shift = y - center_y
    r = np.sqrt(x_shift * x_shift + y_shift * y_shift)
    mask = np.zeros_like(x, dtype="int")
    mask[(r >= r_in) & (r < r_out)] = 1
    return mask



__author__ = "Maverick-Oh"

import numpy as np
import pytest
import numpy.testing as npt
import lenstronomy.Util.param_util as param_util
from lenstronomy.Util import util


class TestEPL_BOXYDISKY(object):
    """Test EPL_BOXYDISKY vs EPL + MULTIPOLE values."""

    def setup_method(self):
        from lenstronomy.LensModel.Profiles.epl import EPL

        self.epl = EPL()
        from lenstronomy.LensModel.Profiles.multipole import Multipole

        self.multipole = Multipole()
        from lenstronomy.LensModel.Profiles.epl_boxydisky import EPL_BOXYDISKY

        self.epl_boxydisky = EPL_BOXYDISKY()

        self.x, self.y = util.make_grid(numPix=10, deltapix=0.2)
        self.theta_E_list = [0.5, 1, 2]
        self.gamma_list = [1.8, 2.0, 2.2]
        self.e1_list = [-0.2, 0.0, 0.2]
        self.e2_list = [-0.2, 0.0, 0.2]
        self.a4_a_list = [0.0, 0.05, -0.05]

    def test_function(self):
        for gamma in self.gamma_list:
            for e1 in self.e1_list:
                for e2 in self.e2_list:
                    for theta_E in self.theta_E_list:
                        for a4_a in self.a4_a_list:
                            kwargs_epl = {
                                "theta_E": theta_E,
                                "gamma": gamma,
                                "e1": e1,
                                "e2": e2,
                            }
                            phi, q = param_util.ellipticity2phi_q(e1, e2)
                            kwargs_multipole = {
                                "m": 4,
                                "a_m": a4_a * theta_E / np.sqrt(q),
                                "phi_m": phi,
                            }
                            kwargs_epl_boxydisky = {
                                "theta_E": theta_E,
                                "gamma": gamma,
                                "e1": e1,
                                "e2": e2,
                                "a4_a": a4_a,
                            }
                            value1 = self.epl.function(
                                self.x, self.y, **kwargs_epl
                            ) + self.multipole.function(
                                self.x, self.y, **kwargs_multipole
                            )
                            value2 = self.epl_boxydisky.function(
                                self.x, self.y, **kwargs_epl_boxydisky
                            )
                            npt.assert_almost_equal(value1, value2, decimal=10)

    def test_derivatives(self):
        for gamma in self.gamma_list:
            for e1 in self.e1_list:
                for e2 in self.e2_list:
                    for theta_E in self.theta_E_list:
                        for a4_a in self.a4_a_list:
                            kwargs_epl = {
                                "theta_E": theta_E,
                                "gamma": gamma,
                                "e1": e1,
                                "e2": e2,
                            }
                            phi, q = param_util.ellipticity2phi_q(e1, e2)
                            kwargs_multipole = {
                                "m": 4,
                                "a_m": a4_a * theta_E / np.sqrt(q),
                                "phi_m": phi,
                            }
                            kwargs_epl_boxydisky = {
                                "theta_E": theta_E,
                                "gamma": gamma,
                                "e1": e1,
                                "e2": e2,
                                "a4_a": a4_a,
                            }

                            f_x1, f_y1 = self.epl.derivatives(
                                self.x, self.y, **kwargs_epl
                            )
                            f_x2, f_y2 = self.multipole.derivatives(
                                self.x, self.y, **kwargs_multipole
                            )
                            f_x = f_x1 + f_x2
                            f_y = f_y1 + f_y2
                            f_x_, f_y_ = self.epl_boxydisky.derivatives(
                                self.x, self.y, **kwargs_epl_boxydisky
                            )

                            npt.assert_almost_equal(f_x, f_x_, decimal=10)
                            npt.assert_almost_equal(f_y, f_y_, decimal=10)

    def test_hessian(self):
        for gamma in self.gamma_list:
            for e1 in self.e1_list:
                for e2 in self.e2_list:
                    for theta_E in self.theta_E_list:
                        for a4_a in self.a4_a_list:
                            kwargs_epl = {
                                "theta_E": theta_E,
                                "gamma": gamma,
                                "e1": e1,
                                "e2": e2,
                            }
                            phi, q = param_util.ellipticity2phi_q(e1, e2)
                            kwargs_multipole = {
                                "m": 4,
                                "a_m": a4_a * theta_E / np.sqrt(q),
                                "phi_m": phi,
                            }
                            kwargs_epl_boxydisky = {
                                "theta_E": theta_E,
                                "gamma": gamma,
                                "e1": e1,
                                "e2": e2,
                                "a4_a": a4_a,
                            }
                            f_xx1, f_xy1, f_yx1, f_yy1 = self.epl.hessian(
                                self.x, self.y, **kwargs_epl
                            )
                            f_xx2, f_xy2, f_yx2, f_yy2 = self.multipole.hessian(
                                self.x, self.y, **kwargs_multipole
                            )
                            f_xx = f_xx1 + f_xx2
                            f_xy = f_xy1 + f_xy2
                            f_yx = f_yx1 + f_yx2
                            f_yy = f_yy1 + f_yy2
                            f_xx_, f_xy_, f_yx_, f_yy_ = self.epl_boxydisky.hessian(
                                self.x, self.y, **kwargs_epl_boxydisky
                            )

                            npt.assert_almost_equal(f_xx, f_xx_, decimal=10)
                            npt.assert_almost_equal(f_xy, f_xy_, decimal=10)
                            npt.assert_almost_equal(f_yx, f_yx_, decimal=10)
                            npt.assert_almost_equal(f_yy, f_yy_, decimal=10)


if __name__ == "__main__":
    pytest.main()



import numpy as np
from coolest.template.classes.parameter import PointEstimate
from coolest.template.classes.probabilities import PosteriorStatistics
from lenstronomy.Util.param_util import ellipticity2phi_q, shear_cartesian2polar


def shapelet_amp_lenstronomy_to_coolest(value):
    """Transforms shapelets coefficients from lenstronomy conventions (x following ra,
    to the left) to COOLEST conventions (x to the right)

    :param value: amplitude of the shapelet (float or np.array) in lenstronomy
        conventions
    :return: amplitude of the shapelet (float or np.array) in COOLEST conventions
    """
    if value is None:
        return None
    else:
        n = 0  # this is the shapelet order
        k = 0  # this is the index of a coefficient for a given order
        new_value = []
        for coeff in value:
            if n % 2 == 0:
                if k % 2 == 1:
                    coeff = -coeff
            elif n % 2 == 1:
                if k % 2 == 0:
                    coeff = -coeff
            new_value.append(coeff)
            k += 1
            if k == n + 1:
                n += 1
                k = 0

        return new_value


def radian_lenstronomy_to_degree_coolest(value):
    """
    Transform an angle in radian from lenstronomy (from x to y with x pointing to the left - aka North of East)
    into an angle in degree for the COOLEST (from y to negative x - aka East of North) with folding

    :param value: float, angle in lenstronomy conventions
    :return: float, angle in COOLEST conventions
    """
    coolest_oriented_degree = folding_coolest(radian_lenstronomy_to_degree(value))
    return coolest_oriented_degree


def radian_lenstronomy_to_degree(value):
    """
    Transform an angle in radian from lenstronomy (from x to y with x pointing to the left - aka North of East)
    into an angle in degree for the COOLEST (from y to negative x - aka East of North) without folding in ]-90;90],
    which is COOLEST convention

    :param value: float, angle in lenstronomy conventions
    :return: float, angle almost in COOLEST conventions (without folding)
    """

    lenstro_degree = value * 180.0 / np.pi
    coolest_oriented_degree = lenstro_degree - 90.0
    coolest_oriented_degree *= -1
    return coolest_oriented_degree


def folding_coolest(value):
    """Folds the angle (already in degree with COOLEST East of North convention) into
    COOLEST range, ]-90;90]

    :param value: float, angle almost in COOLEST conventions (without folding in
        ]-90;90])
    :return: float, angle in COOLEST conventions (with folding in ]-90;90])
    """
    coolest_oriented_degree = value
    if type(coolest_oriented_degree) == type(np.array([])):
        for idx, val in enumerate(coolest_oriented_degree):
            if val <= -90.0:
                coolest_oriented_degree[idx] += 180.0
            elif val > 90.0:
                coolest_oriented_degree[idx] -= 180.0
    else:
        if coolest_oriented_degree <= -90:
            coolest_oriented_degree += 180.0
        elif coolest_oriented_degree > 90:
            coolest_oriented_degree -= 180.0
    return coolest_oriented_degree


def e1e2_lenstronomy_to_qphi_coolest(e1, e2):
    """Transform e1,e2 in lenstronomy to q and phi (axis ratio, position angle East-of-
    North)

    :param e1: lenstronomy usual ellipticity parameters
    :param e2: lenstronomy usual ellipticity parameters
    :return: q, phi ; axis ratio and position angle in COOLEST conventions
    """
    angle, q = ellipticity2phi_q(e1, e2)
    phi = radian_lenstronomy_to_degree_coolest(angle)
    return q, phi


def g1g2_lenstronomy_to_gamma_phi_coolest(gamma1, gamma2):
    """Transform gamma1,gamma2 in lenstronomy to gamma_ext and phi_ext (shear strength,
    position angle East-of-North) with folding.

    :param gamma1: lenstronomy usual shear parameters
    :param gamma2: lenstronomy usual shear parameters
    :return: gamma_ext, phi_ext ; shear strenght and shear position angle in COOLEST
        conventions (with folding)
    """
    angle, gamma_ext = shear_cartesian2polar(gamma1, gamma2)
    phi_ext = radian_lenstronomy_to_degree_coolest(angle)
    return gamma_ext, phi_ext


def g1g2_lenstronomy_to_gamma_phi(gamma1, gamma2):
    """Transform gamma1,gamma2 in lenstronomy to gamma_ext and phi_ext (shear strength,
    position angle East-of-North) without folding.

    :param gamma1: lenstronomy usual shear parameters
    :param gamma2: lenstronomy usual shear parameters
    :return: gamma_ext, phi_ext ; shear strenght and shear position angle almost in
        COOLEST conventions (without folding)
    """

    angle = np.arctan2(gamma2, gamma1) / 2.0
    phi_ext = radian_lenstronomy_to_degree(angle)

    gamma_ext = np.sqrt(gamma1**2 + gamma2**2)

    return gamma_ext, phi_ext


def shear_update(shear_idx, kwargs_lens, kwargs_lens_mcmc=None):
    """
    Update the COOLEST SHEAR mass model (gamma_ext - phi_ext) with results in kwargs_lens

    :param shear_idx: coolest.template.classes.profiles.mass.ExternalShear object
    :param kwargs_lens: dictionnary with the point estimate

    :return: updated shear_idx
    """
    gamma_ext, phi_ext = g1g2_lenstronomy_to_gamma_phi_coolest(
        float(kwargs_lens["gamma1"]), float(kwargs_lens["gamma2"])
    )
    shear_idx.parameters["gamma_ext"].set_point_estimate(
        PointEstimate(float(gamma_ext))
    )
    shear_idx.parameters["phi_ext"].set_point_estimate(PointEstimate(float(phi_ext)))

    if kwargs_lens_mcmc is not None:
        g1 = [arg["gamma1"] for arg in kwargs_lens_mcmc]
        g2 = [arg["gamma2"] for arg in kwargs_lens_mcmc]

        g_ext, p_ext = g1g2_lenstronomy_to_gamma_phi(np.array(g1), np.array(g2))

        g_ext_mean = np.mean(g_ext)
        g_ext_16, g_ext_50, g_ext_84 = np.quantile(g_ext, [0.16, 0.5, 0.84])
        shear_idx.parameters["gamma_ext"].set_posterior(
            PosteriorStatistics(
                float(g_ext_mean), float(g_ext_50), float(g_ext_16), float(g_ext_84)
            )
        )

        p_ext_mean = folding_coolest(np.mean(p_ext))
        p_ext_16, p_ext_50, p_ext_84 = folding_coolest(
            np.quantile(p_ext, [0.16, 0.5, 0.84])
        )
        shear_idx.parameters["phi_ext"].set_posterior(
            PosteriorStatistics(
                float(p_ext_mean), float(p_ext_50), float(p_ext_16), float(p_ext_84)
            )
        )
    print("shear correctly updated")

    return


def pemd_update(mass, kwargs_lens, kwargs_lens_mcmc=None):
    """Update the COOLEST PEMD mass model with results in kwargs_lens.

    :param mass: coolest.template.classes.profiles.mass.PEMD object
    :param kwargs_lens: dictionnary with the point estimate
    :return: updated mass
    """
    q, phi = e1e2_lenstronomy_to_qphi_coolest(
        float(kwargs_lens["e1"]), float(kwargs_lens["e2"])
    )
    mass.parameters["theta_E"].set_point_estimate(
        PointEstimate(float(kwargs_lens["theta_E"]))
    )
    mass.parameters["gamma"].set_point_estimate(
        PointEstimate(float(kwargs_lens["gamma"]))
    )
    mass.parameters["q"].set_point_estimate(PointEstimate(float(q)))
    mass.parameters["phi"].set_point_estimate(PointEstimate(float(phi)))
    mass.parameters["center_x"].set_point_estimate(
        PointEstimate(-float(kwargs_lens["center_x"]))
    )
    mass.parameters["center_y"].set_point_estimate(
        PointEstimate(float(kwargs_lens["center_y"]))
    )

    if kwargs_lens_mcmc is not None:
        te = [arg["theta_E"] for arg in kwargs_lens_mcmc]
        te_mean = np.mean(te)
        te_16, te_50, te_84 = np.quantile(te, [0.16, 0.5, 0.84])
        mass.parameters["theta_E"].set_posterior(
            PosteriorStatistics(
                float(te_mean), float(te_50), float(te_16), float(te_84)
            )
        )

        g = [arg["gamma"] for arg in kwargs_lens_mcmc]
        g_mean = np.mean(g)
        g_16, g_50, g_84 = np.quantile(g, [0.16, 0.5, 0.84])
        mass.parameters["gamma"].set_posterior(
            PosteriorStatistics(float(g_mean), float(g_50), float(g_16), float(g_84))
        )

        e1 = [arg["e1"] for arg in kwargs_lens_mcmc]
        e2 = [arg["e2"] for arg in kwargs_lens_mcmc]
        ql, phil = e1e2_lenstronomy_to_qphi_coolest(np.array(e1), np.array(e2))

        ql_mean = np.mean(ql)
        ql_16, ql_50, ql_84 = np.quantile(ql, [0.16, 0.5, 0.84])
        mass.parameters["q"].set_posterior(
            PosteriorStatistics(
                float(ql_mean), float(ql_50), float(ql_16), float(ql_84)
            )
        )

        phil_mean = np.mean(phil)
        phil_16, phil_50, phil_84 = np.quantile(phil, [0.16, 0.5, 0.84])
        mass.parameters["phi"].set_posterior(
            PosteriorStatistics(
                float(phil_mean), float(phil_50), float(phil_16), float(phil_84)
            )
        )

        cx = [arg["center_x"] for arg in kwargs_lens_mcmc]
        cx_mean = np.mean(cx)
        cx_16, cx_50, cx_84 = np.quantile(cx, [0.16, 0.5, 0.84])
        mass.parameters["center_x"].set_posterior(
            PosteriorStatistics(
                -float(cx_mean), -float(cx_50), -float(cx_16), -float(cx_84)
            )
        )

        cy = [arg["center_y"] for arg in kwargs_lens_mcmc]
        cy_mean = np.mean(cy)
        cy_16, cy_50, cy_84 = np.quantile(cy, [0.16, 0.5, 0.84])
        mass.parameters["center_y"].set_posterior(
            PosteriorStatistics(
                float(cy_mean), float(cy_50), float(cy_16), float(cy_84)
            )
        )

    print("PEMD correctly updated")

    return


def sie_update(mass, kwargs_lens, kwargs_lens_mcmc=None):
    """Update the COOLEST SIE mass model with results in kwargs_lens.

    :param mass: coolest.template.classes.profiles.mass.SIE object
    :param kwargs_lens : dictionnary with the point estimate
    :return: updated mass
    """

    q, phi = e1e2_lenstronomy_to_qphi_coolest(
        float(kwargs_lens["e1"]), float(kwargs_lens["e2"])
    )
    mass.parameters["theta_E"].set_point_estimate(
        PointEstimate(float(kwargs_lens["theta_E"]))
    )
    mass.parameters["q"].set_point_estimate(PointEstimate(float(q)))
    mass.parameters["phi"].set_point_estimate(PointEstimate(float(phi)))
    mass.parameters["center_x"].set_point_estimate(
        PointEstimate(-float(kwargs_lens["center_x"]))
    )
    mass.parameters["center_y"].set_point_estimate(
        PointEstimate(float(kwargs_lens["center_y"]))
    )

    if kwargs_lens_mcmc is not None:
        te = [arg["theta_E"] for arg in kwargs_lens_mcmc]
        te_mean = np.mean(te)
        te_16, te_50, te_84 = np.quantile(te, [0.16, 0.5, 0.84])
        mass.parameters["theta_E"].set_posterior(
            PosteriorStatistics(
                float(te_mean), float(te_50), float(te_16), float(te_84)
            )
        )

        e1 = [arg["e1"] for arg in kwargs_lens_mcmc]
        e2 = [arg["e2"] for arg in kwargs_lens_mcmc]
        ql, phil = e1e2_lenstronomy_to_qphi_coolest(np.array(e1), np.array(e2))

        ql_mean = np.mean(ql)
        ql_16, ql_50, ql_84 = np.quantile(ql, [0.16, 0.5, 0.84])
        mass.parameters["q"].set_posterior(
            PosteriorStatistics(
                float(ql_mean), float(ql_50), float(ql_16), float(ql_84)
            )
        )

        phil_mean = np.mean(phil)
        phil_16, phil_50, phil_84 = np.quantile(phil, [0.16, 0.5, 0.84])
        mass.parameters["phi"].set_posterior(
            PosteriorStatistics(
                float(phil_mean), float(phil_50), float(phil_16), float(phil_84)
            )
        )

        cx = [arg["center_x"] for arg in kwargs_lens_mcmc]
        cx_mean = np.mean(cx)
        cx_16, cx_50, cx_84 = np.quantile(cx, [0.16, 0.5, 0.84])
        mass.parameters["center_x"].set_posterior(
            PosteriorStatistics(
                -float(cx_mean), -float(cx_50), -float(cx_16), -float(cx_84)
            )
        )

        cy = [arg["center_y"] for arg in kwargs_lens_mcmc]
        cy_mean = np.mean(cy)
        cy_16, cy_50, cy_84 = np.quantile(cy, [0.16, 0.5, 0.84])
        mass.parameters["center_y"].set_posterior(
            PosteriorStatistics(
                float(cy_mean), float(cy_50), float(cy_16), float(cy_84)
            )
        )

    print("SIE correctly updated")

    return


def sersic_update(light, kwargs_light, kwargs_light_mcmc=None):
    """Update the COOLEST Sersic (ellipse) light model with results in kwargs_light.

    :param light: coolest.template.classes.profiles.light.Sersic object
    :param kwargs_light: dictionnary with the point estimate
    :return: updated light
    """
    q, phi = e1e2_lenstronomy_to_qphi_coolest(
        float(kwargs_light["e1"]), float(kwargs_light["e2"])
    )
    light.parameters["I_eff"].set_point_estimate(
        PointEstimate(float(kwargs_light["amp"]))
    )
    light.parameters["theta_eff"].set_point_estimate(
        PointEstimate(float(kwargs_light["R_sersic"]))
    )
    light.parameters["n"].set_point_estimate(
        PointEstimate(float(kwargs_light["n_sersic"]))
    )
    light.parameters["q"].set_point_estimate(PointEstimate(float(q)))
    light.parameters["phi"].set_point_estimate(PointEstimate(float(phi)))
    light.parameters["center_x"].set_point_estimate(
        PointEstimate(-float(kwargs_light["center_x"]))
    )
    light.parameters["center_y"].set_point_estimate(
        PointEstimate(float(kwargs_light["center_y"]))
    )

    if kwargs_light_mcmc is not None:
        a = [arg["amp"] for arg in kwargs_light_mcmc]
        a_mean = np.mean(a)
        a_16, a_50, a_84 = np.quantile(a, [0.16, 0.5, 0.84])
        light.parameters["I_eff"].set_posterior(
            PosteriorStatistics(float(a_mean), float(a_50), float(a_16), float(a_84))
        )

        rs = [arg["R_sersic"] for arg in kwargs_light_mcmc]
        rs_mean = np.mean(rs)
        rs_16, rs_50, rs_84 = np.quantile(rs, [0.16, 0.5, 0.84])
        light.parameters["theta_eff"].set_posterior(
            PosteriorStatistics(
                float(rs_mean), float(rs_50), float(rs_16), float(rs_84)
            )
        )

        ns = [arg["n_sersic"] for arg in kwargs_light_mcmc]
        ns_mean = np.mean(ns)
        ns_16, ns_50, ns_84 = np.quantile(ns, [0.16, 0.5, 0.84])
        light.parameters["n"].set_posterior(
            PosteriorStatistics(
                float(ns_mean), float(ns_50), float(ns_16), float(ns_84)
            )
        )

        e1 = [arg["e1"] for arg in kwargs_light_mcmc]
        e2 = [arg["e2"] for arg in kwargs_light_mcmc]
        ql, phil = e1e2_lenstronomy_to_qphi_coolest(np.array(e1), np.array(e2))

        ql_mean = np.mean(ql)
        ql_16, ql_50, ql_84 = np.quantile(ql, [0.16, 0.5, 0.84])
        light.parameters["q"].set_posterior(
            PosteriorStatistics(
                float(ql_mean), float(ql_50), float(ql_16), float(ql_84)
            )
        )

        phil_mean = np.mean(phil)
        phil_16, phil_50, phil_84 = np.quantile(phil, [0.16, 0.5, 0.84])
        light.parameters["phi"].set_posterior(
            PosteriorStatistics(
                float(phil_mean), float(phil_50), float(phil_16), float(phil_84)
            )
        )

        cx = [arg["center_x"] for arg in kwargs_light_mcmc]
        cx_mean = np.mean(cx)
        cx_16, cx_50, cx_84 = np.quantile(cx, [0.16, 0.5, 0.84])
        light.parameters["center_x"].set_posterior(
            PosteriorStatistics(
                -float(cx_mean), -float(cx_50), -float(cx_16), -float(cx_84)
            )
        )

        cy = [arg["center_y"] for arg in kwargs_light_mcmc]
        cy_mean = np.mean(cy)
        cy_16, cy_50, cy_84 = np.quantile(cy, [0.16, 0.5, 0.84])
        light.parameters["center_y"].set_posterior(
            PosteriorStatistics(
                float(cy_mean), float(cy_50), float(cy_16), float(cy_84)
            )
        )

    print("Sersic (Ellipse) correctly updated")
    return


def shapelets_update(light, kwargs_light, kwargs_light_mcmc=None):
    """Update the COOLEST Shapelets light model with results in kwargs_light.

    :param light: coolest.template.classes.profiles.light.Shapelets object
    :param kwargs_light: dictionnary with the point estimate
    :return: updated light
    """
    light.parameters["amps"].set_point_estimate(
        PointEstimate(
            shapelet_amp_lenstronomy_to_coolest(np.ndarray.tolist(kwargs_light["amp"]))
        )
    )
    light.parameters["beta"].set_point_estimate(
        PointEstimate(float(kwargs_light["beta"]))
    )
    light.parameters["n_max"].set_point_estimate(
        PointEstimate(int(kwargs_light["n_max"]))
    )
    light.parameters["center_x"].set_point_estimate(
        PointEstimate(-float(kwargs_light["center_x"]))
    )
    light.parameters["center_y"].set_point_estimate(
        PointEstimate(float(kwargs_light["center_y"]))
    )

    if kwargs_light_mcmc is not None:
        a = [arg["amp"] for arg in kwargs_light_mcmc]
        a_mean = np.mean(a, axis=0)
        a_16, a_50, a_84 = np.quantile(a, [0.16, 0.5, 0.84], axis=0)
        light.parameters["amps"].set_posterior(
            PosteriorStatistics(
                shapelet_amp_lenstronomy_to_coolest(np.ndarray.tolist(a_mean)),
                shapelet_amp_lenstronomy_to_coolest(np.ndarray.tolist(a_50)),
                shapelet_amp_lenstronomy_to_coolest(np.ndarray.tolist(a_16)),
                shapelet_amp_lenstronomy_to_coolest(np.ndarray.tolist(a_84)),
            )
        )

        b = [arg["beta"] for arg in kwargs_light_mcmc]
        b_mean = np.mean(b)
        b_16, b_50, b_84 = np.quantile(b, [0.16, 0.5, 0.84])
        light.parameters["beta"].set_posterior(
            PosteriorStatistics(float(b_mean), float(b_50), float(b_16), float(b_84))
        )

        nmax = [arg["n_max"] for arg in kwargs_light_mcmc]
        nmax_mean = np.mean(nmax)
        nmax_16, nmax_50, nmax_84 = np.quantile(nmax, [0.16, 0.5, 0.84])
        light.parameters["n_max"].set_posterior(
            PosteriorStatistics(
                float(nmax_mean), float(nmax_50), float(nmax_16), float(nmax_84)
            )
        )

        cx = [arg["center_x"] for arg in kwargs_light_mcmc]
        cx_mean = np.mean(cx)
        cx_16, cx_50, cx_84 = np.quantile(cx, [0.16, 0.5, 0.84])
        light.parameters["center_x"].set_posterior(
            PosteriorStatistics(
                -float(cx_mean), -float(cx_50), -float(cx_16), -float(cx_84)
            )
        )

        cy = [arg["center_y"] for arg in kwargs_light_mcmc]
        cy_mean = np.mean(cy)
        cy_16, cy_50, cy_84 = np.quantile(cy, [0.16, 0.5, 0.84])
        light.parameters["center_y"].set_posterior(
            PosteriorStatistics(
                float(cy_mean), float(cy_50), float(cy_16), float(cy_84)
            )
        )

    print("Shapelets correctly updated")
    return


def lensed_point_source_update(light, kwargs_ps, kwargs_ps_mcmc=None):
    """Update the COOLEST LensedPS light model with results in kwargs_ps.

    :param light: coolest.template.classes.profiles.light.LensedPS object
    :param kwargs_ps: dictionnary with the point estimate
    :return: updated light
    """
    light.parameters["amps"].set_point_estimate(
        PointEstimate(np.ndarray.tolist(kwargs_ps["point_amp"]))
    )
    light.parameters["ra_list"].set_point_estimate(
        PointEstimate(np.ndarray.tolist(-kwargs_ps["ra_image"]))
    )
    light.parameters["dec_list"].set_point_estimate(
        PointEstimate(np.ndarray.tolist(kwargs_ps["dec_image"]))
    )

    if kwargs_ps_mcmc is not None:
        a = [arg["point_amp"] for arg in kwargs_ps_mcmc]
        a_mean = np.mean(a, axis=0)
        a_16, a_50, a_84 = np.quantile(a, [0.16, 0.5, 0.84], axis=0)
        light.parameters["amps"].set_posterior(
            PosteriorStatistics(
                np.ndarray.tolist(a_mean),
                np.ndarray.tolist(a_50),
                np.ndarray.tolist(a_16),
                np.ndarray.tolist(a_84),
            )
        )

        ra = [arg["ra_image"] for arg in kwargs_ps_mcmc]
        ra_mean = np.mean(ra, axis=0)
        ra_16, ra_50, ra_84 = np.quantile(ra, [0.16, 0.5, 0.84], axis=0)
        light.parameters["ra_list"].set_posterior(
            PosteriorStatistics(
                np.ndarray.tolist(-ra_mean),
                np.ndarray.tolist(-ra_50),
                np.ndarray.tolist(-ra_16),
                np.ndarray.tolist(-ra_84),
            )
        )
        dec = [arg["dec_image"] for arg in kwargs_ps_mcmc]
        dec_mean = np.mean(dec, axis=0)
        dec_16, dec_50, dec_84 = np.quantile(dec, [0.16, 0.5, 0.84], axis=0)
        light.parameters["dec_list"].set_posterior(
            PosteriorStatistics(
                np.ndarray.tolist(dec_mean),
                np.ndarray.tolist(dec_50),
                np.ndarray.tolist(dec_16),
                np.ndarray.tolist(dec_84),
            )
        )

    print("Lensed point source correctly updated")
    return



from lenstronomy.PointSource.Types.unlensed import Unlensed
import pytest
import numpy.testing as npt


class TestUnlensed(object):
    def setup_method(self):
        self.ps = Unlensed()
        self.kwargs = {"point_amp": [2, 1], "ra_image": [0, 1], "dec_image": [1, 0]}

    def test_image_position(self):
        x_img, y_img = self.ps.image_position(self.kwargs)
        npt.assert_almost_equal(x_img, self.kwargs["ra_image"])
        npt.assert_almost_equal(y_img, self.kwargs["dec_image"])

    def test_source_position(self):
        x_src, y_src = self.ps.source_position(self.kwargs, kwargs_lens=None)
        npt.assert_almost_equal(x_src, self.kwargs["ra_image"])
        npt.assert_almost_equal(y_src, self.kwargs["dec_image"])

    def test_image_amplitude(self):
        amp = self.ps.image_amplitude(
            self.kwargs,
            kwargs_lens=None,
            x_pos=None,
            y_pos=None,
            magnification_limit=None,
            kwargs_lens_eqn_solver=None,
        )
        npt.assert_almost_equal(amp, self.kwargs["point_amp"])

    def test_source_amplitude(self):
        amp = self.ps.source_amplitude(self.kwargs, kwargs_lens=None)
        npt.assert_almost_equal(amp, self.kwargs["point_amp"])


if __name__ == "__main__":
    pytest.main()
